# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
# Test string to read

import logging
import os
import pytest
import re
import six
import uuid
from pytest import raises

from ...utils import swallow_logs
from ...tests.utils import assert_in, skip_ssh
from ..base import ResourceManager
from niceman.tests.fixtures import get_docker_fixture
from ...consts import TEST_SSH_DOCKER_DIGEST

# Note: due to skip_ssh right here, it would skip the entire module with
# all the tests here if no ssh testing is requested
setup_ssh = skip_ssh(get_docker_fixture)(
    TEST_SSH_DOCKER_DIGEST,
    portmaps={
        49000: 22
    },
    custom_params={
        'host': 'localhost',
        'user': 'root',
        'port': 49000
    },
    scope='module'
)


def test_setup_ssh(setup_ssh):
    # Rudimentary smoke test for setup_ssh so we have
    # multiple uses for the setup_ssh
    assert 'port' in setup_ssh['custom']
    assert setup_ssh['custom']['host'] == 'localhost'


def test_ssh_class(setup_ssh, resource_test_dir):
    with swallow_logs(new_level=logging.DEBUG) as log:

        # Test connecting to test SSH server.
        # TODO: Add a test using a SSH key pair.
        config = dict(
            name='ssh-test-resource',
            type='ssh',
            **setup_ssh['custom']
        )
        resource = ResourceManager.factory(config)
        updated_config = resource.create()
        config.update(updated_config)
        assert re.match('\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$',
                        resource.id) is not None
        assert resource.status == 'N/A'

        # Test running commands in a resource.
        resource.connect(password='root')
        command = ['apt-get', 'install', '-y', 'bc']
        resource.add_command(command)
        command = ['ls', '/']
        resource.add_command(command)
        resource.execute_command_buffer()
        assert_in("Running command '['apt-get', 'install', '-y', 'bc']'", log.lines)
        assert_in("Running command '['ls', '/']'", log.lines)
        # TODO: Figure out why PY3 logger is not picking up STDOUT from SSH server.
        if six.PY2:
            assert_in("Running command '['apt-get', 'install', '-y', 'bc']'", log.lines)
            assert_in("Running command '['ls', '/']'", log.lines)

        # Test SSHSession methods
        session = resource.get_session()

        assert session.exists('/etc/hosts')
        assert session.exists('/no/such/file') is False

        # Copy this file to /root/test_ssh.py on the docker test container.
        session.put(__file__, 'remote_test_ssh.py')
        assert session.exists('remote_test_ssh.py')

        tmp_path = "{}/{}".format(resource_test_dir, uuid.uuid4().hex)
        session.get('/etc/hosts', tmp_path)
        assert os.path.isfile(tmp_path)

        file_contents = session.read('remote_test_ssh.py')
        file_contents = file_contents.splitlines()
        assert file_contents[8] == '# Test string to read'

        path = '/tmp/{}'.format(str(uuid.uuid4()))
        assert session.isdir(path) is False
        session.mkdir(path)
        assert session.isdir(path)

        path = '/tmp/{}/{}'.format(str(uuid.uuid4()), str(uuid.uuid4()))
        assert session.isdir(path) is False
        session.mkdir(path, parents=True)
        assert session.isdir(path)

        assert not session.isdir('not-a-dir')
        assert not session.isdir('/etc/hosts')

        with raises(NotImplementedError):
            session._execute_command('non-existent-command', cwd='/path')


def test_ssh_resource(setup_ssh):

    config = {
        'name': 'ssh-test-resource',
        'type': 'ssh',
        'host': 'localhost',
        'user': 'root',
        'port': 49000
    }
    resource = ResourceManager.factory(config)
    resource.connect(password='root')

    with pytest.raises(NotImplementedError):
        resource.start()
    with pytest.raises(NotImplementedError):
        resource.stop()

    resource.delete()
    assert resource._connection is None

    # resource.get_session()
    # assert type(resource._transport) == paramiko.Transport



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provide information about available backend parameters.
"""

__docformat__ = 'restructuredtext'

from importlib import import_module
from logging import getLogger

from niceman.dochelpers import exc_str
from niceman.interface.base import Interface
from niceman.resource import Resource
from niceman.resource.base import ResourceManager
from niceman.resource.base import get_resource_backends
from niceman.support.constraints import EnsureStr
from niceman.support.param import Parameter


lgr = getLogger('niceman.api.backend_parameters')


def get_resource_classes(names=None):
    for name in names or ResourceManager._discover_types():
        try:
            module = import_module('niceman.resource.{}'.format(name))
        except ImportError as exc:
            import difflib
            known = ResourceManager._discover_types()
            suggestions = difflib.get_close_matches(name, known)
            lgr.warning(
                "Failed to import resource %s: %s. %s: %s",
                name,
                exc_str(exc),
                "Similar backends" if suggestions else "Known backends",
                ', '.join(suggestions or known))
            continue

        class_name = ''.join([token.capitalize() for token in name.split('_')])
        cls = getattr(module, class_name)
        if issubclass(cls, Resource):
            yield name, cls
        else:
            lgr.debug("Skipping %s.%s because it is not a Resource. "
                      "Consider moving away",
                      module, class_name)


class BackendParameters(Interface):
    """Display available backend parameters.
    """

    _params_ = dict(
        backends=Parameter(
            args=("backends",),
            metavar="BACKEND",
            doc="""Restrict output to this backend.""",
            constraints=EnsureStr(),
            nargs="*")
    )

    @staticmethod
    def __call__(backends=None):
        backends = backends or ResourceManager._discover_types()
        for backend, cls in get_resource_classes(backends):
            param_doc = "\n".join(
                ["  {}: {}".format(p, pdoc)
                 for p, pdoc in sorted(get_resource_backends(cls).items())])
            if param_doc:
                out = "Backend parameters for '{}'\n{}".format(
                    backend, param_doc)
            else:
                out = "No backend parameters for '{}'".format(backend)
            print(out)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging
from mock import patch, MagicMock, call

from ...utils import swallow_logs
from ...tests.utils import assert_in, skip_if_no_docker_engine
from ..base import ResourceManager
from ...support.exceptions import ResourceError
from ...consts import TEST_SSH_DOCKER_DIGEST
from ..docker_container import DockerContainer

from niceman.tests.fixtures import get_docker_fixture

from pytest import raises


setup_ubuntu = get_docker_fixture(
    TEST_SSH_DOCKER_DIGEST,
    scope='module',
    name='niceman-test-ssh-container'
)


def test_dockercontainer_class():

    with patch('docker.Client') as client, \
        patch('dockerpty.start') as dockerpty, \
        swallow_logs(new_level=logging.DEBUG) as log:

        client.return_value = MagicMock(
            containers=lambda all: [
                {
                    'Id': '326b0fdfbf83',
                    'Names': ['/existing-test-resource'],
                    'State': 'running'
                },
                {
                    'Id': '111111111111',
                    'Names': ['/duplicate-resource-name'],
                    'State': 'running'
                },
                {
                    'Id': '222222222222',
                    'Names': ['/duplicate-resource-name'],
                    'State': 'running'
                }
            ],
            pull=lambda repository, stream: [
                b'{ "status" : "status 1", "progress" : "progress 1" }',
                b'{ "status" : "status 2", "progress" : "progress 2" }'
            ],
            create_container=lambda name, image, stdin_open, tty, command: {
                'Id': '18b31b30e3a5'
            },
            exec_inspect=lambda id: { 'ExitCode': 0 },
            exec_start=lambda exec_id, stream: [
                b'stdout line 1',
                b'stdout line 2',
                b'stdout line 3'
            ]
        )

        # Test connecting when a resource doens't exist.
        config = {
            'name': 'non-existent-resource',
            'type': 'docker-container'
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        assert resource.id is None
        assert resource.status is None

        # Test catching exception when multiple resources are found at connection.
        config = {
            'name': 'duplicate-resource-name',
            'type': 'docker-container'
        }
        resource = ResourceManager.factory(config)
        with raises(ResourceError) as ecm:
            resource.connect()
        assert ecm.value.args[0].startswith("Multiple container matches found")

        # Test connecting to an existing resource.
        config = {
            'name': 'existing-test-resource',
            'type': 'docker-container',
            'engine_url': 'tcp://127.0.0.1:2375',
            'seccomp_unconfined': True
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        assert resource.image == 'ubuntu:latest'
        assert resource.engine_url == 'tcp://127.0.0.1:2375'
        assert resource.id == '326b0fdfbf83'
        assert resource.name == 'existing-test-resource'
        assert resource.status == 'running'
        assert resource.type == 'docker-container'

        # Test creating an existing resource and catch the exception.
        try:
            resource.create()
        except Exception as e:
            assert e.args[0] == "Container 'existing-test-resource' (ID 326b0fdfbf83) already exists in Docker"

        # Test creating resource.
        config = {
            'name': 'new-test-resource',
            'type': 'docker-container',
            'engine_url': 'tcp://127.0.0.1:2375'
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        results = resource.create()
        assert results['id'] == '18b31b30e3a5'
        assert results['status'] == 'running'
        assert_in('status 1 progress 1', log.lines)
        assert_in('status 2 progress 2', log.lines)

        # Test running commands in a resource.
        command = ['apt-get', 'install', 'bc']
        resource.add_command(command)
        command = ['apt-get', 'install', 'xeyes']
        resource.add_command(command)
        resource.execute_command_buffer()
        assert_in("Running command '['apt-get', 'install', 'bc']'", log.lines)
        assert_in("Running command '['apt-get', 'install', 'xeyes']'", log.lines)

        # Test starting resource.
        resource.start()
        calls = [
            call().start(container='18b31b30e3a5'),
        ]
        client.assert_has_calls(calls, any_order=True)

        # Test logging into the container.
        with resource.get_session(pty=True):
            pass # we do nothing really
        assert dockerpty.called

        # Test stopping resource.
        resource.stop()
        calls = [
            call().stop(container='18b31b30e3a5'),
        ]
        client.assert_has_calls(calls, any_order=True)

        # Test deleting resource.
        resource.delete()
        calls = [
            call().remove_container({'Id': '18b31b30e3a5'}, force=True),
        ]
        client.assert_has_calls(calls, any_order=True)


def test_setup_ubuntu(setup_ubuntu):
    assert setup_ubuntu['container_id']


@skip_if_no_docker_engine
def test_engine_exits():
    assert DockerContainer.is_engine_running()
    assert not DockerContainer.is_engine_running(base_url='foo')


@skip_if_no_docker_engine
def test_container_exists(setup_ubuntu):
    assert DockerContainer.is_container_running(setup_ubuntu['name'])
    assert not DockerContainer.is_container_running('foo')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from mock import patch
from ..config import ConfigManager
from .utils import ok_, eq_, assert_raises, assert_greater

def test_config_empty():
    # nothing to load
    config = ConfigManager(load_default=False)
    eq_(config._get_file_candidates(), [])
    # nevertheless we should be able to specify variables via env

    assert_raises(ValueError, config.getboolean, 'tests', 'somenonexistingone')
    with patch.dict('os.environ', {'NICEMAN_TESTS_NONETWORK': '1'}):
        config.reload()
        ok_(config.getboolean('tests', 'nonetwork'))
        eq_(config.get_as_dtype('tests', 'nonetwork', int), 1)

def test_config_load():
    config = ConfigManager()
    candidates = config._get_file_candidates()
    assert_greater(len(candidates), 2)  # at least system, user, local
    assert_greater(5, len(candidates))  # but shouldn't be too many



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to create an environment
"""

__docformat__ = 'restructuredtext'

from collections import Mapping

from .base import Interface
import niceman.interface.base # Needed for test patching
from ..support.param import Parameter
from ..support.constraints import EnsureStr
from ..support.exceptions import ResourceError
from ..resource import get_manager
from ..dochelpers import exc_str

from logging import getLogger
lgr = getLogger('niceman.api.create')


def parse_backend_parameters(params):
    """Parse a list of backend parameters.

    Parameters
    ----------
    params : sequence of str or mapping
        For a sequence, each item should have the form "<key>=<value".  If
        `params` is a mapping, it will be returned as is.

    Returns
    -------
    A mapping from backend key to value.
    """
    if isinstance(params, Mapping):
        res = params
    elif params:
        res = dict(p.split("=", 1) for p in params)
    else:
        res = {}
    return res


class Create(Interface):
    """Create a computation environment
    """

    _params_ = dict(
        # specs=Parameter(
        #     args=("-s", "--spec",),
        #     dest="specs",
        #     doc="file with specifications (in supported formats) of"
        #         " an environment where execution was originally executed",
        #     metavar='SPEC',
        #     # nargs="+",
        #     constraints=EnsureStr(),
        #     # TODO:  here we need to elaborate options for sub-parsers to
        #     # provide options, like --no-exec, etc  per each spec
        #     # ACTUALLY this type doesn't work for us since it is --spec SPEC SPEC... TODO
        # ),
        name=Parameter(
            args=("name",),
            metavar="NAME",
            doc="""Name of the resource to create""",
            constraints=EnsureStr(),
        ),
        resource_type=Parameter(
            args=("-t", "--resource-type"),
            doc="""Resource type to create""",
            constraints=EnsureStr(),
        ),
        # TODO: Implement --only-env and --existing.
        # only_env=Parameter(
        #     args=("--only-env",),
        #     doc="only env spec",
        #     nargs="+",
        #     #action="store_true",
        # ),
        # existing=Parameter(
        #     args=("-e", "--existing"),
        #     choices=("fail", "redefine"),
        #     doc="Action to take if name is already known"
        # ),
        backend_parameters=Parameter(
            metavar="PARAM",
            args=("-b", "--backend-parameters"),
            nargs="+",
            doc="""One or more backend parameters in the form KEY=VALUE. Use
            the command `niceman backend-parameters` to see the list of
            available backend parameters."""
        ),
    )

    @staticmethod
    def __call__(name, resource_type, backend_parameters):
        # Load, while possible merging/augmenting sequentially
        # provenance = Provenance.factory(specs)
        #
        # TODO: need to be redone to be able to operate based on a spec
        #  we do want
        #     niceman create --resource_type docker_container --spec analysis.spec
        #  which would choose appropriate base container etc
        #
        # if nothing in cmdline instructed on specific one to use:
        #   resource = Resource.factory(resource_type)
        #   resource_base = resource.guess_base(provenance.distributions)
        #
        # internally it might first just check if base OS could be deduced, so
        # we need helpers like
        #     guess_base_os_spec(distributions)
        # and if none is there, each resource, might provide/use defaults, e.g.
        # a default docker image if there is anaconda used and nothing about base
        # env.
        #
        # if not specs:
        #     specs = question("Enter a spec filename", default="spec.yml")

        from niceman.ui import ui

        if not resource_type:
            resource_type = ui.question(
                "Enter a resource type",
                default="docker-container"
            )
        # if only_env:
        #     raise NotImplementedError

        # TODO: Add ability to clone a resource.

        get_manager().create(name, resource_type,
                             parse_backend_parameters(backend_parameters))
        lgr.info("Created the environment %s", name)

        # TODO: at the end install packages using install and created env
        # if not only_env:
        #     from repronim.api import install
        #     install(provenance, name, resource_id, config)


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Run internal NICEMAN (unit)tests to verify correct operation on the system"""


__docformat__ = 'restructuredtext'


import niceman
from .base import Interface


class Test(Interface):
    """Run internal NICEMAN (unit)tests.

    This can be used to verify correct operation on the system
    """
    @staticmethod
    def __call__():
        niceman.test()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# -*- coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test testing utilities

"""

import os
import pytest
import shutil
import sys
import logging
from mock import patch
from six import PY3
from six import text_type

from operator import itemgetter
from os.path import dirname, normpath, pardir, basename
from os.path import isabs, expandvars, expanduser
from collections import OrderedDict

from ..utils import updated, HashableDict, \
    get_cmd_batch_len, execute_command_batch, \
    cmd_err_filter, join_sequence_of_dicts
from os.path import join as opj, abspath, exists
from ..utils import rotree, swallow_outputs, swallow_logs, setup_exceptionhook, md5sum
from ..utils import getpwd, chpwd
from ..utils import auto_repr
from ..utils import find_files
from ..utils import line_profile
from ..utils import not_supported_on_windows
from ..utils import file_basename
from ..utils import expandpath, is_explicit_path
from ..utils import assure_unicode
from ..utils import any_re_search
from ..utils import unique
from ..utils import partition
from ..utils import make_tempfile
from ..utils import on_windows
from ..utils import _path_
from ..utils import to_unicode
from ..utils import generate_unique_name
from ..utils import PathRoot, is_subpath
from ..utils import parse_semantic_version

from .utils import ok_, eq_, assert_false, assert_equal, assert_true

from .utils import with_tempfile, assert_in, with_tree, to_binarystring, \
    is_unicode, is_binarystring, CommandError
from .utils import assert_cwd_unchanged, skip_if_on_windows
from .utils import assure_dict_from_str, assure_list_from_str
from .utils import ok_generator
from .utils import assert_not_in
from .utils import assert_raises
from .utils import ok_startswith


@with_tempfile(mkdir=True)
def __test_rotree(d):  # TODO: redo without AnnexRepo
    d2 = opj(d, 'd1', 'd2')  # deep nested directory
    f = opj(d2, 'f1')
    os.makedirs(d2)
    with open(f, 'w') as f_:
        f_.write("LOAD")
    with swallow_logs():
        pass # ar = AnnexRepo(d2)
    rotree(d)
    # we shouldn't be able to delete anything UNLESS in "crippled" situation:
    # root, or filesystem is FAT etc
    # Theoretically annex should declare FS as crippled when ran as root, but
    # see http://git-annex.branchable.com/bugs/decides_that_FS_is_crippled_under_cowbuilder___40__symlinks_supported_etc__41__/#comment-60c3cbe2710d6865fb9b7d6e247cd7aa
    # so explicit 'or'
    if not (ar.is_crippled_fs() or (os.getuid() == 0)):
        assert_raises(OSError, os.unlink, f)
        assert_raises(OSError, shutil.rmtree, d)
        # but file should still be accessible
        with open(f) as f_:
            eq_(f_.read(), "LOAD")
    # make it RW
    rotree(d, False)
    os.unlink(f)
    shutil.rmtree(d)


def test_swallow_outputs():
    with swallow_outputs() as cm:
        eq_(cm.out, '')
        sys.stdout.write("out normal")
        sys.stderr.write("out error")
        eq_(cm.out, 'out normal')
        sys.stdout.write(" and more")
        eq_(cm.out, 'out normal and more')  # incremental
        eq_(cm.err, 'out error')
        eq_(cm.err, 'out error')  # the same value if multiple times


def test_swallow_logs():
    lgr = logging.getLogger('niceman')
    with swallow_logs(new_level=9) as cm:
        eq_(cm.out, '')
        lgr.log(8, "very heavy debug")
        eq_(cm.out, '')  # not even visible at level 9
        lgr.log(9, "debug1")
        eq_(cm.out, 'debug1\n')  # not even visible at level 9
        lgr.info("info")
        eq_(cm.out, 'debug1\ninfo\n')  # not even visible at level 9


@pytest.mark.parametrize("interactive", [True, False])
def test_setup_exceptionhook(interactive):
    old_exceptionhook = sys.excepthook

    post_mortem_tb = []

    def our_post_mortem(tb):
        post_mortem_tb.append(tb)

    with patch('sys.excepthook'), \
            patch('niceman.utils.is_interactive', lambda: interactive), \
            patch('pdb.post_mortem', our_post_mortem):
        setup_exceptionhook()
        our_exceptionhook = sys.excepthook
        ok_(old_exceptionhook != our_exceptionhook)
        #out = sys.stdout
        with swallow_logs() as cml, swallow_outputs() as cmo:
            # we need to call our_exceptionhook explicitly b/c nose
            # swallows all Exceptions and hook never gets executed
            try:
                raise RuntimeError
            except Exception as e:  # RuntimeError:
                type_, value_, tb_ = sys.exc_info()
            our_exceptionhook(type_, value_, tb_)
            if PY3:
                # Happens under tox environment but not in manually crafted ones -- not yet sure
                # what it is about but --dbg does work with python3 so lettting it skip for now
                pytest.skip("TODO: Not clear why in PY3 calls cleanup if we try to access the beast")
            assert_in('Traceback (most recent call last)', cmo.err)
            assert_in('in test_setup_exceptionhook', cmo.err)
            if interactive:
                assert_equal(post_mortem_tb[0], tb_)
            else:
                assert_equal(post_mortem_tb, [])
                # assert_in('We cannot setup exception hook', cml.out)

    eq_(old_exceptionhook, sys.excepthook)


def test_md5sum():
    # just a smoke (encoding/decoding) test for md5sum
    _ = md5sum(__file__)


# archives support in with_tree disabled in niceman's copy
# @with_tree([('1.tar.gz', (('1 f.txt', '1 f load'),))])
# def test_md5sum_archive(d):
#     # just a smoke (encoding/decoding) test for md5sum
#     _ = md5sum(opj(d, '1.tar.gz'))

def test_updated():
    d = {}
    eq_(updated(d, {1: 2}), {1: 2})
    eq_(d, {})

    d = {'a': 'b'}
    eq_(updated(d, ((0, 1), (2, 3))), {0: 1, 'a': 'b', 2: 3})
    eq_(d, {'a': 'b'})

    # and that it would maintain the type
    d = OrderedDict(((99, 0), ('z', 0), ('a', 0)))
    d_ = updated(d, {0: 1})
    ok_(isinstance(d_, OrderedDict))
    eq_(d_, OrderedDict(((99, 0), ('z', 0), ('a', 0), (0, 1))))


def test_get_local_file_url_windows():
    pytest.skip("TODO")


@assert_cwd_unchanged
def test_getpwd_basic():
    pwd = getpwd()
    ok_(isabs(pwd))
    eq_(os.getcwd(), abspath(pwd))

    # that we do not chdir anywhere if None provided
    with patch('os.chdir') as oschdir:
        with chpwd(None):
            eq_(getpwd(), pwd)
        assert_false(oschdir.called)


@skip_if_on_windows
@with_tempfile(mkdir=True)
@assert_cwd_unchanged
def test_getpwd_symlink(tdir=None):
    sdir = opj(tdir, 's1')
    pwd_orig = getpwd()
    os.symlink('.', sdir)
    s1dir = opj(sdir, 's1')
    s2dir = opj(sdir, 's2')
    try:
        chpwd(sdir)
        pwd = getpwd()
        eq_(pwd, sdir)
        chpwd('s1')
        eq_(getpwd(), s1dir)
        chpwd('.')
        eq_(getpwd(), s1dir)
        chpwd('..')
        eq_(getpwd(), sdir)
    finally:
        chpwd(pwd_orig)

    # test context handler way of use
    with chpwd(s1dir):
        eq_(getpwd(), s1dir)
    eq_(getpwd(), pwd_orig)

    assert_false(exists(s2dir))
    with pytest.raises(OSError):
        with chpwd(s2dir):
            pass
    with chpwd(s2dir, mkdir=True):
        ok_(exists(s2dir))
        eq_(getpwd(), s2dir)


def test_auto_repr():

    class withoutrepr:
        def __init__(self):
            self.a = "does not matter"

    @auto_repr
    class buga:
        def __init__(self):
            self.a = 1
            self.b = list(range(100))
            self.c = withoutrepr()
            self._c = "protect me"

        def some(self):
            return "some"

    assert_equal(repr(buga()), "buga(a=1, b=<<[0, 1, 2, 3, 4, 5, 6, ...>>, c=<withoutrepr>)")
    assert_equal(buga().some(), "some")


def test_assure_list_from_str():
    assert_equal(assure_list_from_str(''), None)
    assert_equal(assure_list_from_str([]), None)
    assert_equal(assure_list_from_str('somestring'), ['somestring'])
    assert_equal(assure_list_from_str('some\nmultiline\nstring'), ['some', 'multiline', 'string'])
    assert_equal(assure_list_from_str(['something']), ['something'])
    assert_equal(assure_list_from_str(['a', 'listof', 'stuff']), ['a', 'listof', 'stuff'])


def test_assure_dict_from_str():
    assert_equal(assure_dict_from_str(''), None)
    assert_equal(assure_dict_from_str({}), None)
    assert_equal(assure_dict_from_str(
            '__ac_name={user}\n__ac_password={password}\nsubmit=Log in\ncookies_enabled='), dict(
             __ac_name='{user}', __ac_password='{password}', cookies_enabled='', submit='Log in'))
    assert_equal(assure_dict_from_str(
        dict(__ac_name='{user}', __ac_password='{password}', cookies_enabled='', submit='Log in')), dict(
             __ac_name='{user}', __ac_password='{password}', cookies_enabled='', submit='Log in'))


def test_any_re_search():
    assert_true(any_re_search('a', 'a'))
    assert_true(any_re_search('a', 'bab'))
    assert_false(any_re_search('^a', 'bab'))
    assert_true(any_re_search(['b', '.ab'], 'bab'))
    assert_false(any_re_search(['^b', 'bab'], 'ab'))


def test_find_files():
    tests_dir = dirname(__file__)
    proj_dir = normpath(opj(dirname(__file__), pardir))

    ff = find_files('.*', proj_dir)
    ok_generator(ff)
    files = list(ff)
    assert(len(files) > 10)  # we have more than 10 test files here
    assert_in(opj(tests_dir, 'test_utils.py'), files)
    # and no directories should be mentioned
    assert_not_in(tests_dir, files)

    ff2 = find_files('.*', proj_dir, dirs=True)
    files2 = list(ff2)
    assert_in(opj(tests_dir, 'test_utils.py'), files2)
    assert_in(tests_dir, files2)

    # now actually matching the path
    ff3 = find_files('.*/test_.*\.py$', proj_dir, dirs=True)
    files3 = list(ff3)
    assert_in(opj(tests_dir, 'test_utils.py'), files3)
    assert_not_in(tests_dir, files3)
    for f in files3:
        ok_startswith(basename(f), 'test_')

@with_tree(tree={
    '.git': {
        '1': '2'
    },
    'd1': {
        '.git': 'possibly a link from submodule'
    },
    'git': 'just a file'
})
def test_find_files_exclude_vcs(repo=None):
    ff = find_files('.*', repo, dirs=True)
    files = list(ff)
    assert_equal({basename(f) for f in files}, {'d1', 'git'})
    assert_not_in(opj(repo, '.git'), files)

    ff = find_files('.*', repo, dirs=True, exclude_vcs=False)
    files = list(ff)
    assert_equal({basename(f) for f in files}, {'d1', 'git', '.git', '1'})
    assert_in(opj(repo, '.git'), files)


def test_not_supported_on_windows():
    with patch('niceman.utils.on_windows', True):
        assert_raises(NotImplementedError, not_supported_on_windows)
        assert_raises(NotImplementedError, not_supported_on_windows, "msg")

    with patch('niceman.utils.on_windows', False):
        assert_equal(not_supported_on_windows(), None)
        assert_equal(not_supported_on_windows("msg"), None)


def test_file_basename():
    eq_(file_basename('1'), '1')
    eq_(file_basename('d1/1'), '1')
    eq_(file_basename('/d1/1'), '1')
    eq_(file_basename('1.'), '1.')
    eq_(file_basename('1.tar.gz'), '1')
    eq_(file_basename('1.Tar.gz'), '1')
    eq_(file_basename('1._bak.gz'), '1')
    eq_(file_basename('1.tar.gz', return_ext=True), ('1', 'tar.gz'))
    eq_(file_basename('/tmp/1.tar.gz'), '1')
    eq_(file_basename('/tmp/1.longish.gz'), '1.longish')
    eq_(file_basename('1_R1.1.1.tar.gz'), '1_R1.1.1')
    eq_(file_basename('ds202_R1.1.1.tgz'), 'ds202_R1.1.1')


def test_expandpath():
    eq_(expandpath("some", False), expanduser('some'))
    eq_(expandpath("some", False), expandvars('some'))
    assert_true(isabs(expandpath('some')))
    # this may have to go because of platform issues
    eq_(expandpath("$HOME"), expanduser('~'))


def test_is_explicit_path():
    # by default expanded paths are absolute, hence explicit
    assert_true(is_explicit_path(expandpath('~')))
    assert_false(is_explicit_path("here"))


def test_make_tempfile():
    # check if mkdir, content conflict caught
    with pytest.raises(ValueError):
        with make_tempfile(content="blah", mkdir=True):  # pragma: no cover
            pass


def test_unique():
    eq_(unique(range(3)), [0, 1, 2])
    eq_(unique((1, 0, 1, 3, 2, 0, 1)), [1, 0, 3, 2])
    eq_(unique([]), [])
    eq_(unique([(1, 2), (1,), (1, 2), (0, 3)]), [(1, 2), (1,), (0, 3)])

    # with a key now
    eq_(unique([(1, 2), (1,), (1, 2), (0, 3)], key=itemgetter(0)), [(1, 2), (0, 3)])
    eq_(unique([(1, 2), (1, 3), (1, 2), (0, 3)], key=itemgetter(1)), [(1, 2), (1, 3)])


def test_partition():
    def fn(*args, **kwargs):
        left, right = partition(*args, **kwargs)
        return list(left), list(right)

    eq_(fn([False, True, False]),
        ([False, False], [True]))

    eq_(fn([1, 5, 4, 10], lambda x: x > 4),
        ([1, 4], [5, 10]))

    eq_(fn([1, 5, 4, 10], lambda x: x < 0),
        ([1, 5, 4, 10], []))


def test_path_():
    eq_(_path_('a'), 'a')
    if on_windows:
        eq_(_path_('a/b'), r'a\b')
    else:
        p = 'a/b/c'
        assert(_path_(p) is p)  # nothing is done to it whatsoever


def test_assure_unicode():
    ok_(isinstance(assure_unicode("m"), text_type))
    ok_(isinstance(assure_unicode('grandchild_äöü東'), text_type))
    ok_(isinstance(assure_unicode(u'grandchild_äöü東'), text_type))
    eq_(assure_unicode('grandchild_äöü東'), u'grandchild_äöü東')
    # now, non-utf8
    # Decoding could be deduced with high confidence when the string is
    # really encoded in that codepage
    mom_koi8r = u"мама".encode('koi8-r')
    eq_(assure_unicode(mom_koi8r), u"мама")
    eq_(assure_unicode(mom_koi8r, confidence=0.9), u"мама")
    mom_iso8859 = u'mamá'.encode('iso-8859-1')
    eq_(assure_unicode(mom_iso8859), u'mamá')
    eq_(assure_unicode(mom_iso8859, confidence=0.5), u'mamá')
    # but when we mix, it does still guess something allowing to decode:
    mixedin = mom_koi8r + u'東'.encode('iso2022_jp') + u'東'.encode('utf-8')
    ok_(isinstance(assure_unicode(mixedin), text_type))
    # but should fail if we request high confidence result:
    with assert_raises(ValueError):
        assure_unicode(mixedin, confidence=0.9)
    # For other, non string values, actually just returns original value
    # TODO: RF to actually "assure" or fail??  For now hardcoding that assumption
    assert assure_unicode(1) is 1


def test_unicode_and_binary_conversion():
    s = "Test String"
    s_unicode = to_unicode(s)
    s_binary = to_binarystring(s)
    assert (is_unicode(s_unicode))
    assert (not(is_binarystring(s_unicode)))
    assert (is_binarystring(s_binary))
    assert (not(is_unicode(s_binary)))


def test_generate_unique_set():
    names = set()
    n = generate_unique_name("test_%d", names)
    assert(n == "test_0")
    names.add(n)
    n = generate_unique_name("test_%d", names)
    assert(n == "test_1")


def test_hashable_dict():
    key_a = HashableDict({"a": 1, "b": "test"})
    key_b = HashableDict({"a": 1, "b": "test"})
    key_c = HashableDict({"a": "dog", "b": "boo"})
    d = dict()
    d[key_a] = 1
    assert(key_b in d)
    assert(key_c not in d)


def test_cmd_err_filter():
    my_filter = cmd_err_filter("testing")
    assert my_filter(CommandError("", "", None, "", "testing"))
    assert not my_filter(CommandError("", "", None, "", "failure"))
    assert not my_filter(ValueError("not CommandError"))


def test_join_sequence_of_dicts():
    assert join_sequence_of_dicts(({"a": 1, "b": 2}, {"c": 3}, {"d": 4})) == \
           {"a": 1, "b": 2, "c": 3, "d": 4}
    with pytest.raises(RuntimeError):
        join_sequence_of_dicts(({"a": 1, "b": 2}, {"b": 3}, {"d": 4}))


def test_get_cmd_batch_len_empty():
    with pytest.raises(ValueError) as cm:
        get_cmd_batch_len([], 10)
    cm.match("Cannot batch")


def test_execute_command_batch():
    # Create a dummy session that can possibly raise a ValueError
    class DummySession(object):
        def execute_command(self, cmd):
            if cmd[0] == "ValueError":
                raise ValueError
            else:
                return (str(len(cmd)), None)
    session = DummySession()
    # First let's do a simple test to count the args
    args = list(map(str, range(1, 101)))
    cmd_gen = execute_command_batch(session, [], args, None)
    for (out, _, _) in cmd_gen:
        assert out == "100"
    # Now let's raise an exception but not list it as handled
    cmd_gen = execute_command_batch(session, ["ValueError"], args, None)
    with pytest.raises(ValueError):
        list(cmd_gen)
    # Now let's raise an exception
    cmd_gen = execute_command_batch(session, ["ValueError"], args,
                                    lambda x: isinstance(x, ValueError))
    for (_, _, err) in cmd_gen:
        assert isinstance(err, ValueError)


def test_pathroot():
    proot = PathRoot(lambda s: s.endswith("root"))
    assert proot("") is None
    assert proot("/") is None
    assert proot("/not_a_r_oot") is None
    assert proot("/root") == "/root"
    assert proot("/root/a_root_it_is_not") == "/root"
    assert proot("/root/x/child_root") == "/root/x/child_root"


def test_is_subpath(tmpdir):
    tmpdir = str(tmpdir)

    assert is_subpath("a/b", "a")
    assert not is_subpath("b", "a/b")

    # Partial matches in the parent directory are not false positives.
    assert not is_subpath("abc/b", "a")

    assert is_subpath("/tmp/a", "/tmp")
    assert is_subpath("/tmp/a/b/c", "/tmp")
    assert is_subpath("/tmp/a/b/c", "/tmp")
    assert not is_subpath("/tmp", "/tmp/a/b/c")
    # Same path is considered a "subpath".
    assert is_subpath("/tmp", "/tmp")
    # Trailing slashes don't matter.
    assert is_subpath("/tmp/", "/tmp")


def test_parse_semantic_version():
    for version, expected in [("1.2.3", ("1", "2", "3", "")),
                              ("12.2.33", ("12", "2", "33", "")),
                              ("1.2.3.rc1", ("1", "2", "3", ".rc1")),
                              ("1.2.3-blah", ("1", "2", "3", "-blah"))]:
        assert parse_semantic_version(version) == expected

    with pytest.raises(ValueError):
        parse_semantic_version("X.Y.Z")
    with pytest.raises(ValueError):
        parse_semantic_version("1.2")


def test_line_profile():
    pytest.importorskip("line_profiler")

    @line_profile
    def f(j):
        i = j + 1  # xyz
        return i

    with swallow_outputs() as cmo:
        assert_equal(f(3), 4)
        assert_equal(cmo.err, '')
        assert_in('i = j + 1  # xyz', cmo.out)

# NOTE: test_line_profile must be the last one in the file
#       since line_profiler obscures the coverage reports.
#       So add any new test above it


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from ..reprozip import ReprozipProvenance
from .constants import REPROZIP_SPEC2_YML_FILENAME


def test_load_config():
    config = ReprozipProvenance(REPROZIP_SPEC2_YML_FILENAME)
    files_all = config.get_files()
    files_noother = config.get_files(limit='other')
    assert len(files_noother) < len(files_all)
    # TODO: more testing




# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from niceman.cmdline.main import main

import functools
import uuid
from mock import patch
import os
import os.path as op
import pytest

import attr

from niceman.api import execute
from niceman.formats.niceman import NicemanProvenance
from niceman.utils import swallow_outputs
from niceman.interface.execute import TracedCommand
from niceman.support.external_versions import external_versions
from ...resource.base import ResourceManager
from ...tests.utils import assert_is_subset_recur
from ...tests.utils import skip_if_no_apt_cache
from ...tests.utils import skip_if_no_network
from ...tests.utils import skip_ssh
from ...tests.fixtures import get_docker_fixture
from ...consts import TEST_SSH_DOCKER_DIGEST


docker_container = skip_ssh(get_docker_fixture)(
    TEST_SSH_DOCKER_DIGEST,
    name='testing-container',
    scope='module',
    seccomp_unconfined=True
)


def test_execute_interface(docker_container):

    with patch('niceman.resource.ResourceManager._get_inventory') as get_inventory:
        config = {
            "status": "running",
            "engine_url": "unix:///var/run/docker.sock",
            "type": "docker-container",
            "name": "testing-container",
        }

        get_inventory.return_value = {
            "testing-container": config
        }

        for internal in [False, True]:
            path = '/tmp/{}'.format(str(uuid.uuid4()))
            cmd = ['execute', '--resource', 'testing-container']
            if internal:
                cmd.append('--internal')
            cmd.extend(['mkdir', path])

            manager = ResourceManager()
            with patch("niceman.interface.execute.get_manager",
                       return_value=manager):
                main(cmd)

                session = manager.get_resource("testing-container").get_session()
                assert session.exists(path)

                # on 2nd run mkdir should fail since already exists
                with swallow_outputs() as cmo:
                    with pytest.raises(SystemExit) as cme:
                        main(cmd)
                    assert cme.value.code == 1
                    assert "File exists" in cmo.err


def test_invalid_trace_internal():
    with pytest.raises(RuntimeError):
        execute("doesn't matter", [], internal=True, trace=True)


@pytest.fixture(scope="function")
def trace_info(tmpdir_factory):
    """Return a TracedCommand that uses temporary directories.
    """
    remote_dir = str(tmpdir_factory.mktemp("remote"))
    local_dir = str(tmpdir_factory.mktemp("local"))
    cls = functools.partial(TracedCommand,
                            remote_dir=remote_dir, local_dir=local_dir)
    return {"remote": remote_dir,
            "local": local_dir,
            "class": cls}


def test_trace_docker(docker_container, trace_info):
    with patch("niceman.resource.ResourceManager._get_inventory") as get_inv:
        config = {"status": "running",
                  "engine_url": "unix:///var/run/docker.sock",
                  "type": "docker-container",
                  "name": "testing-container"}
        get_inv.return_value = {"testing-container": config}
        manager = ResourceManager()
        with patch("niceman.interface.execute.get_manager",
                   return_value=ResourceManager()):
            with patch("niceman.interface.execute.CMD_CLASSES",
                       {"trace": trace_info["class"]}):
                execute("ls", ["-l"],
                        trace=True, resref="testing-container")
            # Barely more than a smoke test.  The test_trace_docker will look
            # at the generated spec.
            session = manager.get_resource("testing-container").get_session()
            assert session.exists(trace_info["remote"])
            # The tracer didn't doing anything with the local test directory:
            assert not os.listdir(trace_info["remote"])


@pytest.mark.integration
# Avoiding skip_if_no_network and skip_if_no_apt_cache because it leads to a
# TypeError under Python 2.
@pytest.mark.skipif(os.environ.get('NICEMAN_TESTS_NONETWORK'),
                    reason="No network")
@pytest.mark.skipif("cmd:apt-cache" not in external_versions,
                    reason="No apt-cache")
def test_trace_local(trace_info):
    with patch("niceman.resource.ResourceManager._get_inventory") as get_inv:
        config = {"status": "running",
                  "type": "shell",
                  "name": "testing-local"}
        get_inv.return_value = {"testing-local": config}
        with patch("niceman.interface.execute.get_manager",
                   return_value=ResourceManager()):
            with patch("niceman.interface.execute.CMD_CLASSES",
                       {"trace": trace_info["class"]}):
                execute("ls", ["-l"], trace=True, resref="testing-local")

    local_dir = trace_info["local"]
    assert set(os.listdir(local_dir)) == {"traces", "tracers"}
    trace_dirs = os.listdir(op.join(local_dir, "traces"))
    assert len(trace_dirs) == 1

    prov = NicemanProvenance(op.join(local_dir, "traces",
                                     trace_dirs[0], "niceman.yml"))
    deb_dists = [dist for dist in prov.get_distributions()
                 if dist.name == "debian"]
    assert len(deb_dists) == 1

    expect = {"packages": [{"files": ["/bin/ls"], "name": "coreutils"}]}
    assert_is_subset_recur(expect, attr.asdict(deb_dists[0]), [dict, list])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for cmdline.helpers"""

__docformat__ = 'restructuredtext'

from ..helpers import strip_arg_from_argv

from ...tests.utils import eq_


def test_strip_arg_from_argv():
    eq_(strip_arg_from_argv(['-s', 'value'], 'value', ('-s',)), [])
    eq_(strip_arg_from_argv(['-s', 'value'], 'value', ('-s', '--long-s')), [])
    eq_(strip_arg_from_argv(
            ['cmd', '-s', 'value', '--more'], 'value', ('-s', '--long-s')),
            ['cmd',                '--more'])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Utilities for working with pip.
"""
import itertools
import json
import os
import re

from niceman.utils import execute_command_batch


def parse_pip_show(out):
    pip_info = {}
    list_tag = None
    for line in out.splitlines():
        if line.startswith("#"):   # Skip if comment
            continue
        if line.startswith("  "):  # List item
            item = line[2:].strip()
            if list_tag and item:  # Add the item to the existing list
                pip_info[list_tag].append(item)
            continue
        if ":" in line:            # List tag or tag/value
            split_line = line.split(":", 1)
            tag = split_line[0].strip()
            value = None
            if len(split_line) > 1:  # Parse the value if there
                value = split_line[1].strip()
            if value:                # We have both a tag and a value
                pip_info[tag] = value
                list_tag = None      # A new tag stops the previous list
            else:                    # We have just a list_tag so start it
                list_tag = tag
                pip_info[list_tag] = []

    return pip_info


def _pip_batched_show(session, which_pip, pkgs):
    cmd = [which_pip, "show", "-f"]
    batch = execute_command_batch(session, cmd, pkgs)
    sep_re = re.compile("^---$", flags=re.MULTILINE)
    entries = (sep_re.split(stacked) for stacked, _, _ in batch)

    for pkg, entry in zip(pkgs, itertools.chain(*entries)):
        info = parse_pip_show(entry)
        yield pkg, info


def pip_show(session, which_pip, pkgs):
    """Gather package details from `pip show`.

    Parameters
    ----------
    session : Session instance
        Session in which to execute the command.
    which_pip : str
        Name of the pip executable.
    pkgs : sequence
        Collection of packages pass to the command.

    Returns
    -------
    A tuple of two dicts, where the first maps a package name to its
    details, and the second maps package files to the package name.
    """
    packages = {}
    file_to_pkg = {}

    show_entries = _pip_batched_show(session, which_pip, pkgs)

    for pkg, info in show_entries:
        details = {"name": info["Name"],
                   "version": info["Version"],
                   "location": info["Location"]}
        packages[pkg] = details
        for path in info["Files"]:
            full_path = os.path.normpath(
                os.path.join(info["Location"], path))
            file_to_pkg[full_path] = pkg
    return packages, file_to_pkg


def get_pip_packages(session, which_pip, restriction=None):
    """Return a list of pip packages.

    Parameters
    ----------
    session : Session instance
        Session in which to execute the command.
    which_pip : str
        Name of the pip executable.
    restriction : {None, 'local', 'editable'}, optional
        If 'local', excluded globally installed packages (which pip has access
        to if "--system-site-packages" was used when creating the virtualenv
        directory). If 'editable', only include editable packages.

    Returns
    -------
    A generator that yields package names.
    """
    # We could use either 'pip list' or 'pip freeze' to get a list
    # of packages.  The choice to use 'list' rather than 'freeze'
    # is based on how they show editable packages.  'list' outputs
    # a source directory of the package, whereas 'freeze' outputs
    # a URL like "-e git+https://github.com/[...]".
    cmd = [which_pip, "list", "--format=json"]
    if restriction in ["local", "editable"]:
        cmd.append("--{}".format(restriction))
    out, _ = session.execute_command(cmd)
    return (p["name"] for p in json.loads(out))


def get_package_details(session, which_pip, packages=None):
    """Get package details from `pip show` and `pip list`.

    This is similar to `pip_show`, but it uses `pip list` to get information
    about editable locations and to optionally generate the list of packages.

    Parameters
    ----------
    session : Session instance
        Session in which to execute the command.
    which_pip : str
        Name of the pip executable.
    packages : list of str, optional
        Package names.  If not given, all packages returned by `pip list` are
        used.

    Returns
    -------
    A tuple of two dicts, where the first maps a package name to its
    details and the second maps package files to the package name.
    """
    if packages is None:
        packages = list(get_pip_packages(session, which_pip))
    editable_packages = set(
        get_pip_packages(session, which_pip, restriction="editable"))
    details, file_to_pkg = pip_show(session, which_pip, packages)

    for pkg in details:
        details[pkg]["editable"] = pkg in editable_packages
    return details, file_to_pkg



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test command call wrapper
"""

from mock import patch
import os
import sys
import logging
import shlex
import pytest

from .utils import ok_, eq_, assert_is, assert_equal, assert_false, \
    assert_true, assert_in

from ..cmd import Runner, link_file_load
from ..support.exceptions import CommandError
from ..support.protocol import DryRunProtocol
from .utils import with_tempfile, assert_cwd_unchanged, \
    swallow_outputs, swallow_logs, \
    on_linux, on_osx, on_windows


@assert_cwd_unchanged
@with_tempfile
def test_runner_dry(tempfile=None):

    dry = DryRunProtocol()
    runner = Runner(protocol=dry)

    # test dry command call
    cmd = 'echo Testing dry run > %s' % tempfile
    with swallow_logs(new_level=logging.DEBUG) as cml:
        ret = runner.run(cmd)
        assert_equal(cml.out.rstrip(), "{DryRunProtocol} Running: %s" % cmd)
    assert_equal(("DRY", "DRY"), ret,
                 "Output of dry run (%s): %s" % (cmd, ret))
    assert_equal(shlex.split(cmd, posix=not on_windows), dry[0]['command'])
    assert_false(os.path.exists(tempfile))

    # test dry python function call
    output = runner.call(os.path.join, 'foo', 'bar')
    assert_is(None, output, "Dry call of: os.path.join, 'foo', 'bar' "
                            "returned: %s" % output)
    assert_in('join', dry[1]['command'][0])
    assert_equal("args=('foo', 'bar')", dry[1]['command'][1])


@assert_cwd_unchanged
@with_tempfile
def test_runner(tempfile=None):

    # test non-dry command call
    runner = Runner()
    cmd = 'echo Testing real run > %s' % tempfile
    ret = runner.run(cmd)
    assert_true(os.path.exists(tempfile),
                "Run of: %s resulted with non-existing file %s" %
                (cmd, tempfile))

    # test non-dry python function call
    output = runner.call(os.path.join, 'foo', 'bar')
    assert_equal(os.path.join('foo', 'bar'), output,
                 "Call of: os.path.join, 'foo', 'bar' returned %s" % output)


def test_runner_instance_callable_dry():

    cmd_ = ['echo', 'Testing', '__call__', 'with', 'string']
    for cmd in [cmd_, ' '.join(cmd_)]:
        dry = DryRunProtocol()
        runner = Runner(protocol=dry)
        ret = runner(cmd)
        # (stdout, stderr) is returned.  But in dry -- ("DRY","DRY")
        eq_(ret, ("DRY", "DRY"))
        assert_equal(cmd_, dry[0]['command'],
                     "Dry run of Runner.__call__ didn't record command: %s.\n"
                     "Buffer: %s" % (cmd, dry))

    ret = runner(os.path.join, 'foo', 'bar')
    eq_(ret, None)

    assert_in('join', dry[1]['command'][0],
              "Dry run of Runner.__call__ didn't record function join()."
              "Buffer: %s" % dry)
    assert_equal("args=('foo', 'bar')", dry[1]['command'][1],
                 "Dry run of Runner.__call__ didn't record function join()."
                 "Buffer: %s" % dry)


def test_runner_instance_callable_wet():

    runner = Runner()
    cmd = [sys.executable, "-c", "print('Testing')"]

    out = runner(cmd)
    eq_(out[0].rstrip(), ('Testing'))
    eq_(out[1], '')

    ret = runner(os.path.join, 'foo', 'bar')
    eq_(ret, os.path.join('foo', 'bar'))


def test_runner_log_stderr():
    # TODO: no idea of how to check correct logging via any kind of
    # assertion yet.

    runner = Runner()
    cmd = 'echo stderr-Message should be logged >&2'
    ret = runner.run(cmd, log_stderr=True, expect_stderr=True)

    cmd = 'echo stderr-Message should not be logged >&2'
    with swallow_outputs() as cmo:
        with swallow_logs(new_level=logging.INFO) as cml:
            ret = runner.run(cmd, log_stderr=False)
            eq_(cmo.err.rstrip(), "stderr-Message should not be logged")
            eq_(cml.out, "")


def test_runner_log_stdout():
    # TODO: no idea of how to check correct logging via any kind of
    # assertion yet.

    runner = Runner()
    cmd_ = ['echo', 'stdout-Message should be logged']
    for cmd in [cmd_, ' '.join(cmd_)]:
        # should be identical runs, either as a string or as a list
        kw = {}
        # on Windows it can't find echo if ran outside the shell
        if on_windows and isinstance(cmd, list):
            kw['shell'] = True
        with swallow_logs(logging.DEBUG) as cm:
            ret = runner.run(cmd, log_stdout=True, **kw)
            eq_(cm.lines[0], "Running: %s" % cmd)
            if not on_windows:
                # we can just count on sanity
                eq_(cm.lines[1], "stdout| stdout-"
                                 "Message should be logged")
            else:
                # echo outputs quoted lines for some reason, so relax check
                ok_("stdout-Message should be logged" in cm.lines[1])

    cmd = 'echo stdout-Message should not be logged'
    with swallow_outputs() as cmo:
        with swallow_logs(new_level=logging.INFO) as cml:
            ret = runner.run(cmd, log_stdout=False)
            eq_(cmo.out, "stdout-Message should not be logged\n")
            eq_(cml.out, "")


@with_tempfile
def test_link_file_load(tempfile=None):
    tempfile2 = tempfile + '_'

    with open(tempfile, 'w') as f:
        f.write("LOAD")

    link_file_load(tempfile, tempfile2)  # this should work in general

    ok_(os.path.exists(tempfile2))

    with open(tempfile2, 'r') as f:
        assert_equal(f.read(), "LOAD")

    def inode(fname):
        with open(fname) as fd:
            return os.fstat(fd.fileno()).st_ino

    def stats(fname, times=True):
        """Return stats on the file which should have been preserved"""
        with open(fname) as fd:
            st = os.fstat(fd.fileno())
            stats = (st.st_mode, st.st_uid, st.st_gid, st.st_size)
            if times:
                return stats + (st.st_atime, st.st_mtime)
            else:
                return stats
            # despite copystat mtime is not copied. TODO
            #        st.st_mtime)

    if on_linux or on_osx:
        # above call should result in the hardlink
        assert_equal(inode(tempfile), inode(tempfile2))
        assert_equal(stats(tempfile), stats(tempfile2))

        # and if we mock absence of .link
        def raise_AttributeError(*args):
            raise AttributeError("TEST")

        with patch('os.link', raise_AttributeError):
            with swallow_logs(logging.WARNING) as cm:
                link_file_load(tempfile, tempfile2)  # should still work
                ok_("failed (TEST), copying file" in cm.out)

    # should be a copy (either originally for windows, or after mocked call)
    ok_(inode(tempfile) != inode(tempfile2))
    with open(tempfile2, 'r') as f:
        assert_equal(f.read(), "LOAD")
    assert_equal(stats(tempfile, times=False), stats(tempfile2, times=False))
    os.unlink(tempfile2)  # TODO: next two with_tempfile


@with_tempfile(mkdir=True)
def test_runner_failure(dir_=None):
    runner = Runner()
    failing_cmd = ['sh', '-c', 'exit 2']

    with swallow_logs() as cml:
        with pytest.raises(CommandError) as cme:
            runner.run(failing_cmd, cwd=dir_)
        assert_in('Failed to run', cml.out)
        assert_equal(2, cme.value.code)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging

from os import linesep

from ... import __version__
from ...dochelpers import exc_str
from ...version import __version__
from ..external_versions import ExternalVersions, LooseVersion
from ...tests.utils import assert_true, assert_false
from ...tests.utils import assert_equal, assert_greater_equal, assert_greater
from ..exceptions import CommandError
from ..exceptions import OutdatedExternalDependency, MissingExternalDependency
from ...tests.utils import (
    with_tempfile,
    create_tree,
    swallow_logs,
)

import pytest
from mock import patch
from six import PY3

if PY3:
    # just to ease testing
    def cmp(a, b):
        return (a > b) - (a < b)


def test_external_versions_basic():
    ev = ExternalVersions()
    our_module = 'niceman'
    assert_equal(ev.versions, {})
    assert_equal(ev[our_module], __version__)
    # and it could be compared
    assert_greater_equal(ev[our_module], __version__)
    assert_greater(ev[our_module], '0.0.0a1')
    assert_equal(list(ev.keys()), [our_module])
    assert_true(our_module in ev)
    assert_false('unknown' in ev)

    # all are LooseVersions now
    assert_true(isinstance(ev[our_module], LooseVersion))
    version_str = __version__
    assert_equal(ev.dumps(), "Versions: %s=%s" % (our_module, version_str))

    # For non-existing one we get None
    assert_equal(ev['custom__nonexisting'], None)
    # and nothing gets added to _versions for nonexisting
    assert_equal(set(ev.versions.keys()), {our_module})

    # but if it is a module without version, we get it set to UNKNOWN
    assert_equal(ev['os'], ev.UNKNOWN)
    # And get a record on that inside
    assert_equal(ev.versions.get('os'), ev.UNKNOWN)
    # And that thing is "True", i.e. present
    assert(ev['os'])
    # but not comparable with anything besides itself (was above)
    pytest.raises(TypeError, cmp, ev['os'], '0')
    pytest.raises(TypeError, assert_greater, ev['os'], '0')

    return
    # Code below is from original duecredit, and we don't care about
    # testing this one
    # And we can get versions based on modules themselves
    from niceman.tests import mod
    assert_equal(ev[mod], mod.__version__)

    # Check that we can get a copy of the versions
    versions_dict = ev.versions
    versions_dict[our_module] = "0.0.1"
    assert_equal(versions_dict[our_module], "0.0.1")
    assert_equal(ev[our_module], __version__)


def test_external_version_contains():
    ev = ExternalVersions()
    assert_true("niceman" in ev)
    assert_false("does not exist" in ev)


def test_external_versions_unknown():
    assert_equal(str(ExternalVersions.UNKNOWN), 'UNKNOWN')


def test_external_versions_smoke():
    ev = ExternalVersions()
    assert_false(linesep in ev.dumps())
    assert_true(ev.dumps(indent=True).endswith(linesep))


@pytest.mark.parametrize("modname",
                         ['scipy', 'numpy', 'mvpa2', 'sklearn', 'statsmodels',
                          'pandas', 'matplotlib', 'psychopy'])
def test_external_versions_popular_packages(modname):
    ev = ExternalVersions()
    try:
        exec("import %s" % modname, globals(), locals())
    except ImportError:
        pytest.skip("External %s not present" % modname)
    except Exception as e:
        pytest.skip("External %s fails to import: %s" % (modname, exc_str(e)))
    assert (ev[modname] is not ev.UNKNOWN)
    assert_greater(ev[modname], '0.0.1')
    assert_greater('1000000.0', ev[modname])  # unlikely in our lifetimes


def test_external_versions_rogue_module(tmpdir):
    topd = str(tmpdir)
    ev = ExternalVersions()
    # if module throws some other non-ImportError exception upon import
    # we must not crash, but issue a warning
    modname = 'verycustomrogue__'
    create_tree(topd, {modname + '.py': 'raise Exception("pickaboo")'})
    with patch('sys.path', [topd]), \
        swallow_logs(new_level=logging.WARNING) as cml:
        assert ev[modname] is None
        assert_true(ev.dumps(indent=True).endswith(linesep))
        assert 'pickaboo' in cml.out


def test_custom_versions():
    ev = ExternalVersions()
    ev.CUSTOM = {'bogus': lambda: 1 / 0}
    assert_equal(ev['bogus'], None)


class thing_with_tuple_version:
    __version__ = (0, 1)


class thing_with_list_version:
    __version__ = [0, 1]


@pytest.mark.parametrize("thing",
                         [thing_with_tuple_version, thing_with_list_version,
                          '0.1', (0, 1), [0, 1]])
def test_list_tuple(thing):
    version = ExternalVersions._deduce_version(thing)
    assert_greater(version, '0.0.1')
    assert_greater('0.2', version)
    assert_equal('0.1', version)
    assert_equal(version, '0.1')


def test_humanize():
    # doesn't provide __version__
    assert ExternalVersions()['humanize']


def test_check():
    ev = ExternalVersions()
    # should be all good
    ev.check('niceman')
    ev.check('niceman', min_version=__version__)

    with pytest.raises(MissingExternalDependency):
        ev.check('dataladkukaracha')
    with pytest.raises(MissingExternalDependency) as cme:
        ev.check('dataladkukaracha', min_version="buga", msg="duga")

    assert "duga" in str(cme.value)

    with pytest.raises(OutdatedExternalDependency):
        ev.check('niceman', min_version="10000000")  # we will never get there!



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NICEMAN package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
'''Unit tests for Python API functionality.'''

import re
from niceman.utils import getargspec
import pytest

from .utils import assert_true, assert_false, eq_


def test_basic_setup():
    # the import alone will verify that all default values match their
    # constraints
    from niceman import api
    # random pick of something that should be there
    assert_true(hasattr(api, 'create'))
    assert_true(hasattr(api, 'test'))
    # make sure all helper utilities do not pollute the namespace
    # and we end up only with __...__ attributes
    assert_false(list(filter(lambda s: s.startswith('_') and not re.match('__.*__', s), dir(api))))


def get_interface_specs():
    from importlib import import_module
    from niceman.interface.base import get_interface_groups

    for grp_name, grp_descr, interfaces in get_interface_groups():
        for intfspec in interfaces:
            # turn the interface spec into an instance
            mod = import_module(intfspec[0], package='niceman')
            intf = getattr(mod, intfspec[1])
            spec = getattr(intf, '_params_', dict())

            # figure out which of the specs are "positional"
            spec_posargs = {
                name
                for name, param in spec.items()
                if param.cmd_args and not param.cmd_args[0].startswith('-')
            }
            # we have information about positional args
            yield intf, spec_posargs


interface_specs = list(get_interface_specs())


@pytest.mark.parametrize("intf,spec_posargs", interface_specs,
                         ids=[x[0].__name__ for x in interface_specs])
def test_consistent_order_of_args(intf, spec_posargs):
    f = getattr(intf, '__call__')
    args, varargs, varkw, defaults = getargspec(f)
    # now verify that those spec_posargs are first among args
    if not spec_posargs:
        pytest.skip("no positional args")
    eq_(set(args[:len(spec_posargs)]), spec_posargs)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Environment sub-class to provide management of a Singularity environment."""

import attr
import os
from six.moves import shlex_quote

from ..cmd import Runner
from ..dochelpers import borrowdoc
from ..support.exceptions import CommandError
from .session import POSIXSession, Session
from .base import Resource
from ..utils import attrib
from ..utils import command_as_string

import logging
lgr = logging.getLogger('niceman.resource.singularity')


@attr.s
class Singularity(Resource):
    """
    Environment manager which manages a Singularity container.
    """

    # Generic properties of any Resource
    name = attrib()

    # Container properties
    id = attrib()
    image = attrib(doc="Base image filename or url")
    type = attrib(default='singularity')

    status = attrib()
    _runner = attrib()

    def connect(self):
        """
        Open a connection to the environment.
        """
        self._runner = Runner()

        # Make sure singularity is installed
        stdout, _ = self._runner.run(['singularity', '--version'])
        if stdout.startswith('2.2') or stdout.startswith('2.3'):
            # Running singularity instances and managing them didn't happen
            # until version 2.4. See: https://singularity.lbl.gov/archive/
            raise CommandError(msg="Singularity version >= 2.4 required.")

        # Get instance info if we have one running.
        info = self.get_instance_info()
        if info:
            if not self.name:
                self.name = info['name']
            if not self.id:
                self.id = "{name}-{pid}".format(**info)
            self.image = info['image']
            self.status = 'running'
        else:
            self.id = None
            self.status = None

    def create(self):
        """
        Create a container instance.

        Returns
        -------
        dict : config parameters to capture in the inventory file
        """
        # Check to see if the container is already running.
        info = self.get_instance_info()
        if info:
            lgr.info('Resource {} already exists.'.format(self.name))
        else:
            # Start the container instance.
            # NOTE: Logging stdout and stderr hangs the run call, so we
            # disable the logging in the run call below.
            self._runner.run(['singularity', 'instance.start', self.image,
                self.name], log_stdout=False, log_stderr=False)
            info = self.get_instance_info()

        # Update status
        self.id = "{name}-{pid}".format(**info)
        self.status = 'running'
        return {
            'id': self.id,
            'status': self.status
        }

    def delete(self):
        """
        Deletes a container instance.
        """
        # Check to see if the container is already stopped.
        if not self.get_instance_info():
            return

        # Stop the container.
        self._runner.run(['singularity', 'instance.stop', self.name])

        # Update status
        self.id = None
        self.status = None

    def start(self):
        """
        Start a stopped container.
        """
        # Not a Singularity feature
        raise NotImplementedError

    def stop(self):
        """
        Stop a running container.
        """
        # Not a Singularity feature
        raise NotImplementedError

    @borrowdoc(Resource)
    def get_session(self, pty=False, shared=None):
        if not self.get_instance_info():
            self.connect()

        if pty and shared is not None and not shared:
            lgr.warning("Cannot do non-shared pty session for Singularity yet")
        return (PTYSingularitySession if pty else SingularitySession)(
            name=self.name
        )

    def get_instance_info(self):
        """
        Return basic information about a running container

        Returns
        -------
        dict : instance info
        """
        try:
            stdout, _ = self._runner.run(['singularity', 'instance.list'],
                expect_fail=True)
        except CommandError:
            return None

        # Parse stdout to find the running instance.
        # The output of instance.list is a table with columns:
        #   DAEMON NAME, PID, and CONTAINER IMAGE
        # Daemon names are unique on each server.
        for row in stdout.splitlines()[1:]:
            items = row.split()
            if self.name == items[0] or self.id == "{0}-{1}".format(*items):
                return {
                    'name': items[0],  # daemon name
                    'pid': items[1],   # daemon process id
                    'image': items[2]  # container image file
                }


@attr.s
class SingularitySession(POSIXSession):
    name = attrib(default=attr.NOTHING)
    _runner = Runner()

    @borrowdoc(Session)
    def _execute_command(self, command, env=None, cwd=None):
        if env:
            raise NotImplementedError("passing env variables to session execution")
        if cwd:
            raise NotImplementedError("handle cwd for singularity")
        lgr.debug('Running command %r', command)
        stdout, stderr = self._runner.run(
            "singularity exec instance://{} {}".format(
                self.name, command_as_string(command)),
            expect_fail=True)

        return (stdout, stderr)

    @borrowdoc(Session)
    def put(self, src_path, dest_path, uid=-1, gid=-1):
        dest_path = self._prepare_dest_path(src_path, dest_path,
                                            local=False, absolute_only=True)
        cmd = 'cat {} | singularity exec instance://{} tee {} > /dev/null'
        self._runner.run(cmd.format(shlex_quote(src_path),
                                    self.name,
                                    shlex_quote(dest_path)))

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid)

    @borrowdoc(Session)
    def get(self, src_path, dest_path=None, uid=-1, gid=-1):
        dest_path = self._prepare_dest_path(src_path, dest_path)
        cmd = 'singularity exec instance://{} cat {} > {}'
        self._runner.run(cmd.format(self.name,
                                    shlex_quote(src_path),
                                    shlex_quote(dest_path)))

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid, remote=False)


@attr.s
class PTYSingularitySession(SingularitySession):
    """Interactive Singularity Session"""

    @borrowdoc(Session)
    def open(self):
        lgr.debug("Opening TTY connection to singularity container.")
        cmdline = ['singularity', 'shell', 'instance://' + self.name]
        # TODO: Until we work out how the Niceman session interactive prompt is
        # going to work, we are returning to the OS command prompt when the
        # resource login shell closes. The following line will change after the
        # session management code is developed.
        os.execlp('singularity', *cmdline)

    @borrowdoc(Session)
    def close(self):
        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NICEMAN package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Base classes for UI

"""

__docformat__ = 'restructuredtext'

import abc

from six import add_metaclass

from ..utils import auto_repr


@auto_repr
@add_metaclass(abc.ABCMeta)
class InteractiveUI(object):
    """Semi-abstract class for interfaces to implement interactive UI"""

    @abc.abstractmethod
    def question(self, text,
                 title=None, choices=None,
                 default=None,
                 error_message=None,
                 hidden=False):
        pass

    def yesno(self, *args, **kwargs):
        default = kwargs.pop('default', None)
        if default:
            if hasattr(default, 'lower'):
                default = default.lower()
            elif isinstance(default, bool):
                default = {True: 'yes', False: 'no'}[default]
        response = self.question(
            *args, choices=['yes', 'no', 'y', 'n'],
            default=default,
            **kwargs
        ).rstrip('\n')
        if response in ('yes', 'y'):
            return True
        elif response in ('no', 'n'):
            return False
        else:
            raise RuntimeError("must not happen but did")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from niceman.cmdline.main import main

import logging
from mock import patch, call, MagicMock

from ...resource.base import ResourceManager
from ...utils import swallow_logs
from ...tests.utils import assert_in


def test_install_interface(demo1_spec):

    with patch('docker.Client') as client, \
        patch('niceman.distributions.debian.DebianDistribution.install_packages'), \
        patch('niceman.resource.ResourceManager._get_inventory') as get_inventory, \
        patch('requests.get') as requests, \
        swallow_logs(new_level=logging.DEBUG) as log:

        client.return_value = MagicMock(
            containers=lambda all: [
                {
                    'Id': '326b0fdfbf838',
                    'Names': ['/my-resource'],
                    'State': 'running'
                }
            ],
            exec_inspect=lambda id: { 'ExitCode': 0 }
        )

        get_inventory.return_value = {
            "my-resource": {
                "status": "running",
                "engine_url": "tcp://127.0.0.1:2375",
                "type": "docker-container",
                "name": "my-resource",
                "id": "326b0fdfbf838"
            }
        }

        requests.return_value = type("TestObject", (object,), {})()
        requests.return_value.text = '<a href="/archive/debian/20171208T032012Z/dists/sid/">next change</a>'

        args = ['install',
                'my-resource',
                demo1_spec,

        ]
        with patch("niceman.interface.install.get_manager",
                   return_value=ResourceManager()):
            main(args)

        def container_call(cmd):
            return call().exec_create(
                cmd=cmd,
                container={'State': 'running', 'Id': '326b0fdfbf838', 'Names': ['/my-resource']}
            )

        calls = [
            call(base_url='tcp://127.0.0.1:2375'),
            call().exec_create(cmd=['bash', '-c', 'test -e /etc/apt/sources.list.d/niceman.sources.list && echo Found'], container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd='sh -c \'echo "# Niceman repo sources" > /etc/apt/sources.list.d/niceman.sources.list\'', container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot.debian.org/archive/debian/20170531T084046Z/ sid main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot.debian.org/archive/debian/20171208T032012Z/ sid main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ xenial main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ xenial main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd=['apt-key', 'adv', '--recv-keys', '--keyserver', 'hkp://pool.sks-keyservers.net:80', '0xA5D32F012649A5A9'], container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ xenial main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd="grep -q 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ xenial main contrib non-free' /etc/apt/sources.list.d/niceman.sources.list", container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd=['apt-key', 'adv', '--recv-keys', '--keyserver', 'hkp://pool.sks-keyservers.net:80', '0xA5D32F012649A5A9'], container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']}),
            call().exec_create(cmd=['apt-get', '-o', 'Acquire::Check-Valid-Until=false', 'update'], container={'Id': '326b0fdfbf838', 'State': 'running', 'Names': ['/my-resource']})
        ]
        client.assert_has_calls(calls, any_order=True)

        assert_in('Adding Debian update to environment command list.', log.lines)
        assert_in("Running command ['bash', '-c', 'test -e /etc/apt/sources.list.d/niceman.sources.list && echo Found']", log.lines)
        assert_in('Running command "grep -q \'deb http://snapshot.debian.org/archive/debian/20170531T084046Z/ sid main contrib non-free\' /etc/apt/sources.list.d/niceman.sources.list"', log.lines)
        assert_in('Running command "grep -q \'deb http://snapshot.debian.org/archive/debian/20171208T032012Z/ sid main contrib non-free\' /etc/apt/sources.list.d/niceman.sources.list"', log.lines)
        assert_in('Running command "grep -q \'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ xenial main contrib non-free\' /etc/apt/sources.list.d/niceman.sources.list"', log.lines)
        assert_in("Running command ['apt-key', 'adv', '--recv-keys', '--keyserver', 'hkp://pool.sks-keyservers.net:80', '0xA5D32F012649A5A9']", log.lines)
        assert_in("Running command ['apt-get', '-o', 'Acquire::Check-Valid-Until=false', 'update']", log.lines)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface generation

"""

__docformat__ = 'restructuredtext'

import attr
import sys
import re
import textwrap

from ..ui import ui
from logging import getLogger
lgr = getLogger('niceman.interface')


def get_api_name(intfspec):
    """Given an interface specification return an API name for it"""
    if len(intfspec) > 3:
        name = intfspec[3]
    else:
        name = intfspec[0].split('.')[-1]
    return name


def get_cmdline_command_name(intfspec):
    """Given an interface specification return a cmdline command name"""
    if len(intfspec) > 2:
        name = intfspec[2]
    else:
        name = intfspec[0].split('.')[-1].replace('_', '-')
    return name


def get_interface_groups():
    from .. import interface as _interfaces

    grps = []
    # auto detect all available interfaces and generate a function-based
    # API from them
    for _item in _interfaces.__dict__:
        if not _item.startswith('_group_'):
            continue
        grp_name = _item[7:]
        grp = getattr(_interfaces, _item)
        grps.append((grp_name,) + grp)
    return grps


def dedent_docstring(text):
    """Remove uniform indentation from a multiline docstring"""
    # Problem is that first line might often have no offset, so might
    # need to be ignored from dedent call
    if text is None:
        return None
    if not text.startswith(' '):
        lines = text.split('\n')
        if len(lines) == 1:
            # single line, no indentation, nothing to do
            return text
        text2 = '\n'.join(lines[1:])
        return lines[0] + "\n" + textwrap.dedent(text2)
    else:
        return textwrap.dedent(text)


def alter_interface_docs_for_api(docs):
    """Apply modifications to interface docstrings for Python API use."""
    # central place to alter the impression of docstrings,
    # like removing cmdline specific sections
    if not docs:
        return docs
    docs = dedent_docstring(docs)
    # clean cmdline sections
    docs = re.sub(
        '\|\| CMDLINE \>\>.*\<\< CMDLINE \|\|',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # clean cmdline in-line bits
    docs = re.sub(
        '\[CMD:\s[^\[\]]*\sCMD\]',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    docs = re.sub(
        '\[PY:\s([^\[\]]*)\sPY\]',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE)
    docs = re.sub(
        '\|\| PYTHON \>\>(.*)\<\< PYTHON \|\|',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    docs = re.sub(
        '\|\| REFLOW \>\>\n(.*)\<\< REFLOW \|\|',
        lambda match: textwrap.fill(match.group(1)),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    return docs


def alter_interface_docs_for_cmdline(docs):
    """Apply modifications to interface docstrings for cmdline doc use."""
    # central place to alter the impression of docstrings,
    # like removing Python API specific sections, and argument markup
    if not docs:
        return docs
    docs = dedent_docstring(docs)
    # clean cmdline sections
    docs = re.sub(
        '\|\| PYTHON \>\>.*\<\< PYTHON \|\|',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # clean cmdline in-line bits
    docs = re.sub(
        '\[PY:\s[^\[\]]*\sPY\]',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    docs = re.sub(
        '\[CMD:\s([^\[\]]*)\sCMD\]',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE)
    docs = re.sub(
        '\|\| CMDLINE \>\>(.*)\<\< CMDLINE \|\|',
        lambda match: match.group(1),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # remove :role:`...` RST markup for cmdline docs
    docs = re.sub(
        r':\S+:`[^`]*`[\\]*',
        lambda match: ':'.join(match.group(0).split(':')[2:]).strip('`\\'),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # remove None constraint. In general, `None` on the cmdline means don't
    # give option at all, but specifying `None` explicitly is practically
    # impossible
    docs = re.sub(
        ',\sor\svalue\smust\sbe\s`None`',
        '',
        docs,
        flags=re.MULTILINE | re.DOTALL)
    # capitalize variables and remove backticks to uniformize with
    # argparse output
    docs = re.sub(
        '`\S*`',
        lambda match: match.group(0).strip('`').upper(),
        docs)
    # clean up sphinx API refs
    docs = re.sub(
        '\~niceman\.api\.\S*',
        lambda match: "`{0}`".format(match.group(0)[13:]),
        docs)
    # Remove RST paragraph markup
    docs = re.sub(
        r'^.. \S+::',
        lambda match: match.group(0)[3:-2].upper(),
        docs,
        flags=re.MULTILINE)
    docs = re.sub(
        '\|\| REFLOW \>\>\n(.*)\<\< REFLOW \|\|',
        lambda match: textwrap.fill(match.group(1)),
        docs,
        flags=re.MULTILINE | re.DOTALL)
    return docs


def update_docstring_with_parameters(func, params, prefix=None, suffix=None):
    """Generate a useful docstring from a parameter spec

    Amends any existing docstring of a callable with a textual
    description of its parameters. The Parameter spec needs to match
    the number and names of the callables arguments.
    """
    from niceman.utils import getargspec
    # get the signature
    ndefaults = 0
    args, varargs, varkw, defaults = getargspec(func)
    if not defaults is None:
        ndefaults = len(defaults)
    # start documentation with what the callable brings with it
    doc = prefix if prefix else u''
    if len(args) > 1:
        if len(doc):
            doc += '\n'
        doc += "Parameters\n----------\n"
        for i, arg in enumerate(args):
            if arg == 'self':
                continue
            # we need a parameter spec for each argument
            if not arg in params:
                raise ValueError("function has argument '%s' not described as a parameter" % arg)
            param = params[arg]
            # validate the default -- to make sure that the parameter description is
            # somewhat OK
            defaults_idx = ndefaults - len(args) + i
            if defaults_idx >= 0:
                if not param.constraints is None:
                    param.constraints(defaults[defaults_idx])
            orig_docs = param._doc
            param._doc = alter_interface_docs_for_api(param._doc)
            doc += param.get_autodoc(
                arg,
                default=defaults[defaults_idx] if defaults_idx >= 0 else None,
                has_default=defaults_idx >= 0)
            param._doc = orig_docs
            doc += '\n'
    doc += suffix if suffix else u""
    # assign the amended docs
    func.__doc__ = doc
    return func


class Interface(object):
    """Base class for interface implementations"""

    @classmethod
    def setup_parser(cls, parser):
        # XXX needs safety check for name collisions
        # XXX allow for parser kwargs customization
        parser_kwargs = {}
        from niceman.utils import getargspec
        # get the signature
        ndefaults = 0
        args, varargs, varkw, defaults = getargspec(cls.__call__)
        if not defaults is None:
            ndefaults = len(defaults)
        for i, arg in enumerate(args):
            if arg == 'self':
                continue
            param = cls._params_[arg]
            defaults_idx = ndefaults - len(args) + i
            cmd_args = param.cmd_args
            if cmd_args is None:
                cmd_args = []
            if not len(cmd_args):
                if defaults_idx >= 0:
                    # dealing with a kwarg
                    template = '--%s'
                else:
                    # positional arg
                    template = '%s'
                # use parameter name as default argument name
                parser_args = (template % arg.replace('_', '-'),)
            else:
                parser_args = [c.replace('_', '-') for c in cmd_args]
            parser_kwargs = param.cmd_kwargs
            if defaults_idx >= 0:
                parser_kwargs['default'] = defaults[defaults_idx]
            help = alter_interface_docs_for_cmdline(param._doc)
            if help and help[-1] != '.':
                help += '.'
            if param.constraints is not None:
                parser_kwargs['type'] = param.constraints
                # include value constraint description and default
                # into the help string
                cdoc = alter_interface_docs_for_cmdline(
                    param.constraints.long_description())
                if cdoc[0] == '(' and cdoc[-1] == ')':
                    cdoc = cdoc[1:-1]
                help += '  Constraints: %s' % cdoc
            if defaults_idx >= 0:
                help += " [Default: %r]" % (defaults[defaults_idx],)
            # create the parameter, using the constraint instance for type
            # conversion
            parser.add_argument(*parser_args, help=help,
                                **parser_kwargs)

    @classmethod
    def call_from_parser(cls, args):
        # XXX needs safety check for name collisions
        from niceman.utils import getargspec
        argnames = getargspec(cls.__call__)[0]
        kwargs = {k: getattr(args, k) for k in argnames if k != 'self'}
        try:
            return cls.__call__(**kwargs)
        except KeyboardInterrupt:
            ui.error("\nInterrupted by user while doing magic")
            sys.exit(1)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface definition

"""

__docformat__ = 'restructuredtext'


# the following should be series of import definitions for interface implementations
# that shall be exposed in the Python API and the cmdline interface
# all interfaces should be associated with (at least) one of the groups below
_group_dataset = (
    'Commands for manipulating computation environments',
    [
        # source module, source object[, dest. cmdline name[, dest python name]]
        # src module can be relative, but has to be relative to the main 'niceman' package
        ('niceman.interface.create', 'Create'),
        ('niceman.interface.install', 'Install'),
        ('niceman.interface.delete', 'Delete'),
        ('niceman.interface.start', 'Start'),
        ('niceman.interface.stop', 'Stop'),
        ('niceman.interface.login', 'Login'),
        ('niceman.interface.execute', 'Execute'),
        # ('niceman.distribution.run', 'Run'),
    ])

_group_misc = (
    'Miscellaneous commands',
    [
        ('niceman.interface.ls', 'Ls'),
        ('niceman.interface.backend_parameters', 'BackendParameters'),
        # ('niceman.interface.trace', 'Trace'),
        # ('niceman.interface.shell', 'Shell'),
        ('niceman.interface.retrace', 'Retrace'),
        ('niceman.interface.diff', 'Diff'),
        ('niceman.interface.test', 'Test'),
    ])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Plugin support for TriG formatted RDF provenance files.

See: https://en.wikipedia.org/wiki/TriG_(syntax)
"""

from niceman.formats.base import Provenance
from rdflib import ConjunctiveGraph


class TrigProvenance(Provenance):

    @classmethod
    def _load(cls, source):
        graph = ConjunctiveGraph()
        graph.parse(source, format='trig')
        return graph

    # def get_os(self):
    #     return 'Ubuntu'
    #
    # def get_os_version(self):
    #     return '12.04'
    #
    # def get_create_date(self):
    #     return '20091004T111800Z'
    #
    # def get_environment_vars(self):
    #
    #     results = self.graph.query(
    #         """SELECT DISTINCT ?variable ?value
    #         WHERE {
    #         ?x nipype:environmentVariable ?variable .
    #         ?x prov:value ?value .
    #         }""")
    #
    #     return results

    def get_packages(self):

        results = self._src.query(
            """SELECT DISTINCT ?command ?version
            WHERE {
            ?x nipype:command ?full_command .
            bind( strbefore( $full_command, " " ) as ?command ) .
            ?x nipype:version ?version .
            }""")

        return results

    def get_distributions(self):
        # needs to use get_packages and see what is in there --
        # we might need to come up with a "source_distribution" which
        # would just define packages based on their names without clear
        # definition on how they were obtained
        raise NotImplementedError()

    def get_files(self, limit='all'):
        raise NotImplementedError('TODO')


# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test logging facilities """

import re
import os.path
from os.path import exists

from mock import patch

from niceman.log import LoggerHelper

from niceman.tests.utils import with_tempfile, ok_, assert_equal

# pretend we are in interactive mode so we could check if coloring is
# disabled
@patch("niceman.log.is_interactive", lambda: True)
@with_tempfile
def test_logging_to_a_file(dst=None):
    ok_(not exists(dst))

    lgr = LoggerHelper("nicemantest").get_initialized_logger(logtarget=dst)
    ok_(exists(dst))

    msg = "Oh my god, they killed Kenny"
    lgr.error(msg)
    with open(dst) as f:
        lines = f.readlines()
    assert_equal(len(lines), 1, "Read more than a single log line: %s" %  lines)
    line = lines[0]
    ok_(msg in line)
    ok_(not '\033[' in line,
        msg="There should be no color formatting in log files. Got: %s" % line)
    # verify that time stamp and level are present in the log line
    # do not want to rely on not having race conditions around date/time changes
    # so matching just with regexp
    ok_(re.match("\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3} \[ERROR\] %s" % msg,
                 line))

@with_tempfile
def test_logtarget_via_env_variable(dst=None):
    with patch.dict('os.environ', {'NICEMAN_TEST_LOGTARGET': dst}):
        ok_(not exists(dst))
        lgr = LoggerHelper("niceman_test").get_initialized_logger()
        ok_(exists(dst))
    # just to see that mocking patch worked
    ok_(not 'NICEMAN_TEST_LOGTARGET' in os.environ)

@with_tempfile
@with_tempfile
def test_mutliple_targets(dst1=None, dst2=None):
    ok_(not exists(dst1))
    ok_(not exists(dst2))
    lgr = LoggerHelper("nicemantest").get_initialized_logger(
        logtarget="%s,%s" % (dst1, dst2))
    ok_(exists(dst1))
    ok_(exists(dst2))

    msg = "Oh my god, they killed Kenny"
    lgr.error(msg)
    for dst in (dst1, dst2):
        with open(dst) as f:
            lines = f.readlines()
        assert_equal(len(lines), 1, "Read more than a single log line: %s" %  lines)
        ok_(msg in lines[0])

# TODO: somehow test is stdout/stderr get their stuff


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Wrapper for command and function calls, allowing for dry runs and output handling

"""


import subprocess
import sys
import logging
import os
import shutil
import shlex
import atexit
import functools

from six import PY3, PY2
from six import string_types, binary_type
from os.path import abspath, isabs

from .dochelpers import exc_str
from .support.exceptions import CommandError
from .support.protocol import NullProtocol, DryRunProtocol, \
    ExecutionTimeProtocol, ExecutionTimeExternalsProtocol
from .utils import on_windows
from . import cfg

lgr = logging.getLogger('niceman.cmd')

_TEMP_std = sys.stdout, sys.stderr

if PY2:
    # TODO apparently there is a recommended substitution for Python2
    # which is a backported implementation of python3 subprocess
    # https://pypi.python.org/pypi/subprocess32/
    pass

class Runner(object):
    """Provides a wrapper for calling functions and commands.

    An object of this class provides a methods that calls shell commands or
    python functions, allowing for protocolling the calls and output handling.

    Outputs (stdout and stderr) can be either logged or streamed to system's
    stdout/stderr during execution.
    This can be enabled or disabled for both of them independently.
    Additionally, a protocol object can be a used with the Runner. Such a
    protocol has to implement niceman.support.protocol.ProtocolInterface, is
    able to record calls and allows for dry runs.
    """

    __slots__ = ['commands', 'dry', 'cwd', 'env', 'protocol']

    def __init__(self, cwd=None, env=None, protocol=None):
        """
        Parameters
        ----------
        cwd: string, optional
             Base current working directory for commands.  Could be overridden
             per run call via cwd option
        env: dict, optional
             Custom environment to use for calls. Could be overridden per run
             call via env option
        protocol: ProtocolInterface
             Protocol object to write to.
        """

        self.cwd = cwd
        self.env = env
        if protocol is None:
            # TODO: config cmd.protocol = null
            cfg = os.environ.get('NICEMAN_CMD_PROTOCOL', 'null')
            protocol = {
                'externals-time': ExecutionTimeExternalsProtocol,
                'time': ExecutionTimeProtocol,
                'null': NullProtocol
            }[cfg]()
            if cfg != 'null':
                # we need to dump it into a file at the end
                # TODO: config cmd.protocol_prefix = protocol
                filename = '%s-%s.log' % (
                    os.environ.get('NICEMAN_CMD_PROTOCOL_PREFIX', 'protocol'),
                    id(self)
                )
                atexit.register(functools.partial(protocol.write_to_file, filename))

        self.protocol = protocol

    def __call__(self, cmd, *args, **kwargs):
        """Convenience method

        This will call run() or call() depending on the kind of `cmd`.
        If `cmd` is a string it is interpreted as the to be executed command.
        Otherwise it is expected to be a callable.
        Any other argument is passed to the respective method.

        Parameters
        ----------
        cmd: str or callable
           command string to be executed via shell or callable to be called.

        `*args`:
        `**kwargs`:
           see Runner.run() and Runner.call() for available arguments.

        Raises
        ------
        TypeError
          if cmd is neither a string nor a callable.
        """

        if isinstance(cmd, string_types) or isinstance(cmd, list):
            return self.run(cmd, *args, **kwargs)
        elif callable(cmd):
            return self.call(cmd, *args, **kwargs)
        else:
            raise TypeError("Argument 'command' is neither a string, "
                             "nor a list nor a callable.")

    # Two helpers to encapsulate formatting/output
    def _log_out(self, line):
        if line:
            self.log("stdout| " + line.rstrip('\n'))

    def _log_err(self, line, expected=False):
        if line:
            self.log("stderr| " + line.rstrip('\n'),
                     level={True: logging.DEBUG,
                            False: logging.ERROR}[expected])

    def _get_output_online(self, proc, log_stdout, log_stderr,
                           expect_stderr=False, expect_fail=False):
        stdout, stderr = binary_type(), binary_type()
        while proc.poll() is None:
            if log_stdout:
                line = proc.stdout.readline()
                if line:
                    stdout += line
                    self._log_out(line.decode())
                    # TODO: what level to log at? was: level=5
                    # Changes on that should be properly adapted in
                    # test.cmd.test_runner_log_stdout()
            else:
                pass

            if log_stderr:
                line = proc.stderr.readline()
                if line:
                    stderr += line
                    self._log_err(line.decode() if PY3 else line,
                                  expect_stderr or expect_fail)
                    # TODO: what's the proper log level here?
                    # Changes on that should be properly adapted in
                    # test.cmd.test_runner_log_stderr()
            else:
                pass

        return stdout, stderr

    def run(self, cmd, log_stdout=True, log_stderr=True, log_online=False,
            expect_stderr=False, expect_fail=False,
            cwd=None, env=None, shell=None):
        """Runs the command `cmd` using shell.

        In case of dry-mode `cmd` is just added to `commands` and it is
        actually executed otherwise.
        Allows for separately logging stdout and stderr  or streaming it to
        system's stdout or stderr respectively.

        Note: Using a string as `cmd` and shell=True allows for piping,
              multiple commands, etc., but that implies shlex.split() is not
              used. This is considered to be a security hazard.
              So be careful with input.

        Parameters
        ----------
        cmd : str, list
          String (or list) defining the command call.  No shell is used if cmd
          is specified as a list

        log_stdout: bool, optional
            If True, stdout is logged. Goes to sys.stdout otherwise.

        log_stderr: bool, optional
            If True, stderr is logged. Goes to sys.stderr otherwise.

        log_online: bool, optional
            Either to log as output comes in.  Setting to True is preferable
            for running user-invoked actions to provide timely output

        expect_stderr: bool, optional
            Normally, having stderr output is a signal of a problem and thus it
            gets logged at ERROR level.  But some utilities, e.g. wget, use
            stderr for their progress output.  Whenever such output is expected,
            set it to True and output will be logged at DEBUG level unless
            exit status is non-0 (in non-online mode only, in online -- would
            log at DEBUG)

        expect_fail: bool, optional
            Normally, if command exits with non-0 status, it is considered an
            ERROR and logged accordingly.  But if the call intended for checking
            routine, such alarming message should not be logged as ERROR, thus
            it will be logged at DEBUG level.

        cwd : string, optional
            Directory under which run the command (passed to Popen)

        env : string, optional
            Custom environment to pass

        shell: bool, optional
            Run command in a shell.  If not specified, then it runs in a shell
            only if command is specified as a string (not a list)

        Returns
        -------
        (stdout, stderr)

        Raises
        ------
        CommandError
           if command's exitcode wasn't 0 or None. exitcode is passed to
           CommandError's `code`-field. Command's stdout and stderr are stored
           in CommandError's `stdout` and `stderr` fields respectively.
        """

        outputstream = subprocess.PIPE if log_stdout else sys.stdout
        errstream = subprocess.PIPE if log_stderr else sys.stderr

        self.log("Running: %s" % (cmd,))

        if self.protocol.do_execute_ext_commands:

            if shell is None:
                shell = isinstance(cmd, string_types)

            if self.protocol.records_ext_commands:
                prot_exc = None
                prot_id = self.protocol.start_section(
                    shlex.split(cmd, posix=not on_windows)
                    if isinstance(cmd, string_types)
                    else cmd)

            try:
                proc = subprocess.Popen(cmd, stdout=outputstream,
                                        stderr=errstream,
                                        shell=shell,
                                        cwd=cwd or self.cwd,
                                        env=env or self.env)

            except Exception as e:
                prot_exc = e
                lgr.error("Failed to start %r%r: %s" %
                          (cmd, " under %r" % cwd if cwd else '', exc_str(e)))
                raise

            finally:
                if self.protocol.records_ext_commands:
                    self.protocol.end_section(prot_id, prot_exc)

            if log_online:
                out = self._get_output_online(proc, log_stdout, log_stderr,
                                              expect_stderr=expect_stderr,
                                              expect_fail=expect_fail)
            else:
                out = proc.communicate()

            if PY3:
                # Decoding was delayed to this point
                def decode_if_not_None(x):
                    return "" if x is None else binary_type.decode(x)
                # TODO: check if we can avoid PY3 specific here
                out = tuple(map(decode_if_not_None, out))

            status = proc.poll()

            # needs to be done after we know status
            if not log_online:
                self._log_out(out[0])
                if status not in [0, None]:
                    self._log_err(out[1], expected=expect_fail)
                else:
                    # as directed
                    self._log_err(out[1], expected=expect_stderr)

            if status not in [0, None]:
                msg = "Failed to run %r%s. Exit code=%d. out=%s err=%s" \
                    % (cmd, " under %r" % (cwd or self.cwd), status, out[0], out[1])
                (lgr.debug if expect_fail else lgr.error)(msg)
                raise CommandError(str(cmd), msg, status, out[0], out[1])
            else:
                self.log("Finished running %r with status %s" % (cmd, status),
                         level=8)

        else:
            if self.protocol.records_ext_commands:
                self.protocol.add_section(shlex.split(cmd,
                                                      posix=not on_windows)
                                          if isinstance(cmd, string_types)
                                          else cmd, None)
            out = ("DRY", "DRY")

        return out

    def call(self, f, *args, **kwargs):
        """Helper to unify collection of logging all "dry" actions.

        Calls `f` if `Runner`-object is not in dry-mode. Adds `f` along with
        its arguments to `commands` otherwise.

        f : callable

        `*args`, `**kwargs`:
          Callable arguments
        """
        if self.protocol.do_execute_callables:
            if self.protocol.records_callables:
                prot_exc = None
                prot_id = self.protocol.start_section([str(f),
                                                 "args=%s" % str(args),
                                                 "kwargs=%s" % str(kwargs)])

            try:
                return f(*args, **kwargs)
            except Exception as e:
                prot_exc = e
                raise
            finally:
                if self.protocol.records_callables:
                    self.protocol.end_section(prot_id, prot_exc)
        else:
            if self.protocol.records_callables:
                self.protocol.add_section([str(f),
                                             "args=%s" % str(args),
                                             "kwargs=%s" % str(kwargs)],
                                          None)

    def log(self, msg, level=logging.DEBUG):
        """log helper

        Logs at DEBUG-level by default and adds "Protocol:"-prefix in order to
        log the used protocol.
        """
        if isinstance(self.protocol, NullProtocol):
            lgr.log(level, msg)
        else:
            lgr.log(level, "{%s} %s" % (self.protocol.__class__.__name__, msg))


class GitRunner(Runner):
    """
    Runner to be used to run git and git annex commands

    Overloads the runner class to check & update GIT_DIR and
    GIT_WORK_TREE environment variables set to the absolute path
    if is defined and is relative path
    """

    @staticmethod
    def get_git_environ_adjusted(env=None):
        """
        Replaces GIT_DIR and GIT_WORK_TREE with absolute paths if relative path and defined
        """
        git_env = env.copy() if env else os.environ.copy()         # if env set copy else get os environment

        for varstring in ['GIT_DIR', 'GIT_WORK_TREE']:
            var = git_env.get(varstring)
            if var:                                                # if env variable set
                if not isabs(var):                                 # and it's a relative path
                    git_env[varstring] = abspath(var)              # convert it to absolute path
                    lgr.debug("Updated %s to %s" % (varstring, git_env[varstring]))

        return git_env

    def run(self, cmd, env=None, *args, **kwargs):
        return super(GitRunner, self).run(
            cmd, env=self.get_git_environ_adjusted(), *args, **kwargs)

# ####
# Preserve from previous version
# TODO: document intention
# ####
# this one might get under Runner for better output/control
def link_file_load(src, dst, dry_run=False):
    """Just a little helper to hardlink files's load
    """
    dst_dir = os.path.dirname(dst)
    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)
    if os.path.lexists(dst):
        lgr.debug("Destination file %(dst)s exists. Removing it first"
                  % locals())
        # TODO: how would it interact with git/git-annex
        os.unlink(dst)
    lgr.debug("Hardlinking %(src)s under %(dst)s" % locals())
    src_realpath = os.path.realpath(src)

    try:
        os.link(src_realpath, dst)
    except AttributeError as e:
        lgr.warn("Linking of %s failed (%s), copying file" % (src, e))
        shutil.copyfile(src_realpath, dst)
        shutil.copystat(src_realpath, dst)
    else:
        lgr.log(2, "Hardlinking finished")


def get_runner(*args, **kwargs):
    # TODO:  this is all crawl specific -- should be moved away
    if cfg.getboolean('crawl', 'dryrun', default=False):
        kwargs = kwargs.copy()
        kwargs['protocol'] = DryRunProtocol()
    return Runner(*args, **kwargs)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import mock

from niceman.distributions import piputils
from niceman.tests.utils import assert_is_subset_recur


def test_pip_batched_show():
    pkgs = ["pkg0", "pkg1", "pkg2"]
    batches= [("""\
Name: pkg0
Version: 4.1
Home-page: urlwith---three-dashes
Files:
  file0
---
Name: pkg1
Version: 17.4.0
Files:
  file1""",
              None, None),  # err, exception
             ("""\
Name: pkg2
Version: 4""",
              None, None)]

    with mock.patch("niceman.distributions.piputils.execute_command_batch",
                    return_value=batches):
        pkg_entries = list(piputils._pip_batched_show(None, None, pkgs))

    expect = [("pkg0",
               {"Name": "pkg0", "Version": "4.1", "Files": ["file0"],
                # We did not split on the URL's "---".
                "Home-page": "urlwith---three-dashes"}),
              ("pkg1",
               {"Name": "pkg1", "Version": "17.4.0", "Files": ["file1"]}),
              ("pkg2",
               {"Name": "pkg2", "Version": "4"})]
    assert_is_subset_recur(expect, pkg_entries, [dict, list])


def test_parse_pip_show():
    out_base = """\
Name: pkg
Version: 0.3.0
Summary: foo
# Comment line
Home-page: https://github.com/ReproNim/niceman/
Author: marty
Author-email: marty@nowhere
License: GPL3
Location: /local/path/to/site-packages
Requires: six, funcsigs
"""

    # Check parsing of show output for a "standard" package.
    out_files = out_base + """\
Files:
  pkg-0.3.0.dist-info/DESCRIPTION.rst
  pkg/__init__.py"""
    info_files = piputils.parse_pip_show(out_files)

    fields = {"Name", "Version", "Summary", "Home-page", "Author",
              "Author-email", "License", "Location", "Requires", "Files"}

    assert set(info_files.keys()) == fields
    assert info_files["License"] == "GPL3"
    assert info_files["Files"] == ["pkg-0.3.0.dist-info/DESCRIPTION.rst",
                                   "pkg/__init__.py"]

    # Check parsing of show output for an editable packages that lacks
    # files.
    out_no_files = out_base + """\
Files:
Cannot locate installed-files.txt"""
    info_nofiles = piputils.parse_pip_show(out_no_files)
    assert set(info_nofiles.keys()) == fields
    assert info_nofiles["Files"] == []



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
import os
import pytest

import sys
from mock import mock
from subprocess import call

import yaml
import attr

from niceman.formats.niceman import NicemanProvenance
from niceman.tests.utils import create_pymodule
from niceman.tests.utils import skip_if_no_network, assert_is_subset_recur

from niceman.distributions.conda import CondaTracer, CondaDistribution, \
    CondaEnvironment, CondaPackage, CondaChannel, \
    get_conda_platform_from_python, get_miniconda_url


def test_get_conda_platform_from_python():
    assert get_conda_platform_from_python("linux2") == "linux"
    assert get_conda_platform_from_python("darwin") == "osx"


def test_get_miniconda_url():
    assert get_miniconda_url("linux-64", "2.7") == \
           "https://repo.continuum.io/miniconda/Miniconda2-latest-Linux-x86_64.sh"
    assert get_miniconda_url("linux-32", "3.4") == \
           "https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86.sh"
    assert get_miniconda_url("osx-64", "3.5.1") == \
           "https://repo.continuum.io/miniconda/Miniconda3-latest-MacOSX-x86_64.sh"


def test_get_simple_python_version():
    assert CondaDistribution.get_simple_python_version("2.7.12.final.0") == \
           "2.7.12"
    assert CondaDistribution.get_simple_python_version("3.5.1") == \
           "3.5.1"


def test_format_conda_package():
    assert "p" == CondaDistribution.format_conda_package("p")
    assert "p=v" == CondaDistribution.format_conda_package("p", "v")
    assert "p=v=b" == CondaDistribution.format_conda_package("p", "v", "b")
    assert "p" == CondaDistribution.format_conda_package("p", None, "b")


def test_format_pip_package():
    assert "p" == CondaDistribution.format_pip_package("p")
    assert "p==v" == CondaDistribution.format_pip_package("p", "v")


def test_create_conda_export():
    env = CondaEnvironment(
        name="mytest",
        path="/home/butch/.cache/niceman/conda_test/miniconda/envs/mytest",
        packages=[
            CondaPackage(
                name="xz",
                installer=None,
                version="5.2.3",
                build="0",
                channel_name="conda-forge",
                md5="f4e0d30b3caf631be7973cba1cf6f601",
                size="874292",
                url="https://conda.anaconda.org/conda-forge/linux-64/xz-5.2.3-0.tar.bz2",
                files=["bin/xz"]),
            CondaPackage(
                name="rpaths",
                installer="pip",
                version="0.13",
                build=None,
                channel_name=None,
                md5=None,
                size=None,
                url=None,
                files=["lib/python2.7/site-packages/rpaths.py"])],
        channels=[
            CondaChannel(
                name="conda-forge",
                url="https://conda.anaconda.org/conda-forge/linux-64")])
    out = {"name": "mytest",
           "channels": ["conda-forge"],
           "dependencies": ["xz=5.2.3=0",
                            {"pip": ["rpaths==0.13"]}],
           "prefix": "/home/butch/.cache/niceman/conda_test/miniconda/envs/mytest"
           }
    export = yaml.safe_load(CondaDistribution.create_conda_export(env))
    assert export == out


@pytest.mark.integration
@skip_if_no_network
def test_conda_init_install_and_detect():
    test_dir = "/tmp/niceman_conda/miniconda"

    dist = CondaDistribution(
        name="conda",
        path=test_dir,
        conda_version="4.3.31",
        python_version="2.7.14.final.0",
        platform=get_conda_platform_from_python(sys.platform) + "-64",
        environments=[
            CondaEnvironment(
                name="root",
                path=test_dir,
                packages=[
                    CondaPackage(
                        name="conda",
                        installer=None,
                        version="4.3.31",
                        build=None,
                        channel_name=None,
                        md5=None,
                        size=None,
                        url=None,
                        files=None),
                    CondaPackage(
                        name="pip",
                        installer=None,
                        version="9.0.1",
                        build=None,
                        channel_name=None,
                        md5=None,
                        size=None,
                        url=None,
                        files=None),
                    CondaPackage(
                        name="pytest",
                        installer="pip",
                        version="3.4.0",
                        build=None,
                        channel_name=None,
                        md5=None,
                        size=None,
                        url=None,
                        files=None)],
                channels=[
                    CondaChannel(
                        name="conda-forge",
                        url="https://conda.anaconda.org/conda-forge/linux-64"),
                    CondaChannel(
                        name="defaults",
                        url="https://repo.continuum.io/pkgs/main/linux-64")]),
            CondaEnvironment(
                name="mytest",
                path=os.path.join(test_dir, "envs/mytest"),
                packages=[
                    CondaPackage(
                        name="pip",
                        installer=None,
                        version="9.0.1",
                        build=None,
                        channel_name=None,
                        md5=None,
                        size=None,
                        url=None,
                        files=None),
                    CondaPackage(
                        name="xz",
                        installer=None,
                        version="5.2.3",
                        build="0",
                        channel_name="conda-forge",
                        md5="f4e0d30b3caf631be7973cba1cf6f601",
                        size="874292",
                        url="https://conda.anaconda.org/conda-forge/linux-64/xz-5.2.3-0.tar.bz2",
                        files=["bin/xz", ]),
                    CondaPackage(
                        name="rpaths",
                        installer="pip",
                        version="0.13",
                        build=None,
                        channel_name=None,
                        md5=None,
                        size=None,
                        url=None,
                        files=["lib/python2.7/site-packages/rpaths.py"])],
                channels=[
                    CondaChannel(
                        name="conda-forge",
                        url="https://conda.anaconda.org/conda-forge/linux-64")])])
    # First install the environment in /tmp/niceman_conda/miniconda
    dist.initiate(None)
    dist.install_packages()
    # Add an empty environment to test detection of them
    if not os.path.exists(os.path.join(test_dir, "envs/empty")):
        call("cd " + test_dir + "; " +
             "./bin/conda create -y -n empty; ",
             shell=True)

    # Test that editable packages are detected
    pymod_dir = os.path.join(test_dir, "minimal_pymodule")
    if not os.path.exists(pymod_dir):
        create_pymodule(pymod_dir)
        call([os.path.join(test_dir, "envs/mytest/bin/pip"),
              "install", "-e", pymod_dir])

    # Now pick some files we know are in the conda install and detect them
    files = [os.path.join(test_dir, "bin/pip"),
             os.path.join(test_dir, "envs/mytest/bin/xz"),
             os.path.join(test_dir, "envs/empty/conda-meta/history"),
             ]
    tracer = CondaTracer()
    dists = list(tracer.identify_distributions(files))

    assert len(dists) == 1, "Exactly one Conda distribution expected."

    (distributions, unknown_files) = dists[0]

    # NicemanProvenance.write(sys.stdout, distributions)

    assert distributions.platform.startswith(
        get_conda_platform_from_python(sys.platform)), \
        "A conda platform is expected."

    assert len(distributions.environments) == 3, \
        "Three conda environments are expected."

    out = {'environments': [{'name': 'root',
                             'packages': [{'name': 'pip'}]},
                            {'name': 'mytest',
                             'packages': [{'name': 'xz'},
                                          {'name': 'pip'},
                                          {'name': 'rpaths',
                                           'installer': 'pip',
                                           'editable': False},
                                          {'name': 'nmtest',
                                           'files': [],
                                           'installer': 'pip',
                                           'editable': True}]
                             }
                            ]
           }
    assert_is_subset_recur(out, attr.asdict(distributions), [dict, list])

    # conda packages are not repeated as "pip" packages.
    for envs in distributions.environments:
        for pkg in envs.packages:
            if pkg.name == "pip":
                assert pkg.installer is None

    # Smoke test to make sure install_packages doesn't choke on the format that
    # is actually returned by the tracer.
    distributions.initiate(None)
    distributions.install_packages()


def test_get_conda_env_export_exceptions():
    # Mock to capture logs
    def log_warning(msg, *args):
        log_warning.val = msg % args if args else msg

    # Mock to throw unrecognized argument exception
    def raise_unrec_args(_):
        raise Exception("conda-env: error: unrecognized arguments: -p"
                        "/home/butch/old_conda/")

    # Mock to raise some other exception
    def raise_other(_):
        raise Exception("unknown")

    from niceman.distributions.conda import lgr

    tracer = CondaTracer()
    with mock.patch.object(tracer._session, "execute_command",
                           raise_unrec_args), \
         mock.patch.object(lgr, "warning", log_warning):
        tracer._get_conda_env_export("", "/conda")
        assert "Please use Conda 4.3.19" in log_warning.val

    with mock.patch.object(tracer._session, "execute_command",
                           raise_other), \
        mock.patch.object(lgr, "warning", log_warning):
        tracer._get_conda_env_export("", "/conda")
        assert "unknown" in log_warning.val

conda_yaml = os.path.join(os.path.dirname(__file__), 'files', 'conda.yaml')

def test_conda_packages():
    env = NicemanProvenance(conda_yaml).get_environment()
    conda_dist = env.get_distribution(CondaDistribution)
    assert isinstance(conda_dist.packages, list)
    assert len(conda_dist.packages) == 4



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from mock import patch, MagicMock

from ...cmdline.main import main
from ...utils import swallow_logs
from ...resource.base import ResourceManager
from ...tests.utils import assert_in

import logging


def test_ls_interface():
    """
    Test listing the resources.
    """

    with patch('docker.Client') as docker_client, \
        patch('boto3.resource') as aws_client, \
        patch('niceman.resource.ResourceManager._get_inventory') as get_inventory, \
        swallow_logs(new_level=logging.DEBUG) as log:

        docker_client.return_value = MagicMock(
            containers=lambda all: [
                {
                    'Id': '326b0fdfbf83',
                    'Names': ['/my-resource'],
                    'State': 'running'
                }
            ]
        )

        aws_client.side_effect = [
            MagicMock(Instance=lambda id: MagicMock(
                instance_id='i-22221ddf096c22bb0',
                state={'Name': 'running'}
            )),
            MagicMock(Instance=lambda id: MagicMock(
                instance_id='i-3333f40de2b9b8967',
                state={'Name': 'stopped'}
            )),
        ]

        get_inventory.return_value = {
            "docker-resource-1": {
                "status": "running",
                "engine_url": "tcp://127.0.0.1:2375",
                "type": "docker-container",
                "name": "my-resource",
                "id": "326b0fdfbf838"
            },
            "ec2-resource-1": {
                'id': 'i-22221ddf096c22bb0',
                'type': 'aws-ec2',
                'access_key_id': 'my-aws-access-key-id',
                'secret_access_key': 'my-aws-secret-access-key-id',
                'key_name': 'my-ssh-key',
                'key_filename': '/home/me/.ssh/id_rsa',
                "status": "running",
                "name": "aws-resource-1"
            },
            "ec2-resource-2": {
                'id': 'i-3333f40de2b9b8967',
                'type': 'aws-ec2',
                'access_key_id': 'my-aws-access-key-id',
                'secret_access_key': 'my-aws-secret-access-key-id',
                'key_name': 'my-ssh-key',
                'key_filename': '/home/me/.ssh/id_rsa',
                "status": "stopped",
                "name": "aws-resource-2"
            }
        }

        args = [
            'ls',
        ]
        with patch("niceman.interface.login.get_manager",
                   return_value=ResourceManager()):
            main(args)

        assert_in(
            'list result: docker-resource-1, docker-container, 326b0fdfbf838, running',
            log.lines)
        assert_in('list result: ec2-resource-1, aws-ec2, i-22221ddf096c22bb0, running', log.lines)
        assert_in('list result: ec2-resource-2, aws-ec2, i-3333f40de2b9b8967, stopped', log.lines)


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import os
import shutil
import tempfile

import pytest
from .constants import NICEMAN_CFG_PATH
from niceman.cmd import Runner
from niceman.tests.utils import skip_if_no_network, skip_if_no_svn
from niceman.utils import chpwd


def get_docker_fixture(image, portmaps={}, name=None,
                       custom_params={}, scope='function',
                       seccomp_unconfined=False):
    """Produce a fixture which starts/stops a docker container

    It should be called to produce and assign within the scope under some name,
    e.g.::

        setup_ssh = get_docker_fixture(
            scope='module',
            image='rastasheep/ubuntu-sshd:14.04',
            portmaps={
                49000: 22
            }
        )

    to create a `setup_ssh` fixture which would start docker image with port
    mapping having a scope of the module (so starts and stops once, not per each
    test within the module)

    Parameters
    ----------
    image: str
      Docker image to run
    portmaps: dict, optional
      Port mappings (host:container)
    name: str, optional
      Container name
    custom_params: whatever, optional
      What to return in a fixture information in the field 'custom'
    scope: {'function', 'class', 'module', 'session'}, optional
      A scope for the fixture according to `pytest.fixture` docs
    seccomp_unconfined : bool, optional
      Disable kernel secure computing mode when creating the container
    """

    @pytest.fixture(scope=scope)
    def docker_fixture():
        """The actual fixture code generated by get_docker_fixture

        on setup, this fixture ensures that a docker container is running
        and starts one if necessary.

        Fixture yields parameters of the container with a `custom` field passed
        into the `get_docker_container`.

        on teardown, this fixture stops the docker container it started
        """

        skip_if_no_network()
        args = ['docker',
                'run',
                '-d',
                '--rm',
                ]
        if seccomp_unconfined:
            args.extend(['--security-opt', 'seccomp=unconfined'])
        params = {}
        if name:
            args += ['--name', name]
            params['name'] = name

        if portmaps:
            for from_to in portmaps.items():
                args += ['-p', '%d:%d' % from_to]
                params['port'] = from_to[0]
        args += [image]
        stdout, _ = Runner().run(args)
        params['container_id'] = container_id = stdout.strip( )
        params['custom'] = custom_params
        yield params
        Runner().run(['docker', 'stop', container_id])

    return docker_fixture


def git_repo_fixture(kind="default", scope="function"):
    """Create a Git repository fixture.

    Parameters
    ----------
    kind : {"empty", "default", "pair"}, optional
        Kind git repository.

        - empty: a repository with no commits.

        - default: a repository with three commits, each adding one of
          its three files: "foo", "bar", and "subdir/baz".

        - pair: a (local, remote) pair of Git repos.  The repos have
          the same structure as the "default" repo, but "local" has a
          remote "origin" with a URL that points to "remote".

    scope : {"function", "class", "module", "session"}, optional
        A `pytest.fixture` scope argument.

    Returns
    -------
    A fixture function.
    """
    runner = Runner()

    def setup_user():
        # Set the user in the local Git configuration rather than
        # through environmental variables so that test functions using
        # this fixture can make commits under the same user.
        runner(["git", "config", "user.name", "A U Thor"])
        runner(["git", "config", "user.email", "a.thor@example.com"])

    def add_and_commit(fname):
        directory = os.path.dirname(fname)
        if directory and not os.path.exists(directory):
            os.makedirs(directory)
        with open(fname, "w") as fh:
            fh.write(fname + "content")
        runner.run(["git", "add", fname])
        runner.run(["git", "commit", "-m", "add " + fname])

    @pytest.fixture(scope=scope)
    def fixture():
        # We can't use pytest's tempdir because that is limited to
        # scope=function.
        tmpdir = tempfile.mkdtemp(prefix="niceman-tests-")
        repodir = os.path.realpath(os.path.join(tmpdir, "repo0"))
        os.mkdir(repodir)

        retval = repodir

        with chpwd(repodir):
            runner.run(["git", "init"])
            setup_user()

            if kind != "empty":
                add_and_commit("foo")
                add_and_commit("bar")
                runner.run(["git", "tag", "tag0"])
                add_and_commit("subdir/baz")

            if kind == "pair":
                localdir = os.path.realpath(os.path.join(tmpdir, "repo1"))
                runner.run(["git", "clone", repodir, localdir],
                           expect_stderr=True)
                with chpwd(localdir):
                    setup_user()
                retval = localdir, repodir
        yield retval
        shutil.rmtree(tmpdir)
    return fixture


def svn_repo_fixture(kind='default', scope='function'):
    """Create a SVN repository fixture.

    Parameters
    ----------
    kind : {"empty", "default"}

        - empty: a repository with no commits.

        - default: a repository with one commit of file "foo".

    scope : {"function", "class", "module", "session"}, optional
        A `pytest.fixture` scope argument.

    Returns
    -------
    A fixture function.
    """
    @pytest.fixture(scope=scope)
    def fixture():
        skip_if_no_svn()
        repo_name = 'svnrepo'
        tmpdir = os.path.realpath(tempfile.mkdtemp(prefix='niceman-tests-'))
        root_dir = os.path.join(tmpdir, repo_name)
        subdir = os.path.join(tmpdir, 'subdir')
        os.mkdir(subdir)
        runner = Runner()
        runner.run(['svnadmin', 'create', root_dir])
        runner.run(['svn', 'checkout', 'file://' + root_dir], cwd=subdir)
        checked_out_dir = os.path.join(subdir, repo_name)
        if kind != 'empty':
            runner.run(['touch', 'foo'], cwd=checked_out_dir)
            runner.run(['svn', 'add', 'foo'], cwd=checked_out_dir)
            runner.run(['svn', 'commit', '-m', 'bar'], cwd=checked_out_dir)
        yield (root_dir, checked_out_dir)
        shutil.rmtree(tmpdir)
    return fixture
